# pylint: disable=unused-argument, W0621
import csv
import operator
from decimal import ROUND_DOWN
from decimal import Decimal as D

from django.conf import settings
from django.core import exceptions
from django.db import models
from django.db.models.query import Q
from django.template.defaultfilters import date as date_filter
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.timezone import get_current_timezone, now
from django.utils.translation import gettext_lazy as _

from oscar.checks import use_productcategory_materialised_view
from oscar.core.compat import AUTH_USER_MODEL
from oscar.core.loading import cached_import_string, get_class, get_classes, get_model
from oscar.models import fields
from oscar.templatetags.currency_filters import currency

ExpandDownwardsCategoryQueryset = get_class(
    "catalogue.expressions", "ExpandDownwardsCategoryQueryset"
)
ActiveOfferManager, RangeManager, BrowsableRangeManager = get_classes(
    "offer.managers", ["ActiveOfferManager", "RangeManager", "BrowsableRangeManager"]
)
ZERO_DISCOUNT = get_class("offer.results", "ZERO_DISCOUNT")
load_proxy, unit_price = get_classes("offer.utils", ["load_proxy", "unit_price"])


class BaseOfferMixin(models.Model):
    class Meta:
        abstract = True

    def proxy(self):
        """
        Return the proxy model
        """
        klassmap = self.proxy_map
        # Short-circuit logic if current class is already a proxy class.
        if self.__class__ in klassmap.values():
            return self

        field_dict = dict(self.__dict__)
        for field in list(field_dict.keys()):
            if field.startswith("_"):
                del field_dict[field]

        if self.proxy_class:
            klass = load_proxy(self.proxy_class)
            # Short-circuit again.
            if self.__class__ == klass:
                return self
            return klass(**field_dict)
        if self.type in klassmap:
            return klassmap[self.type](**field_dict)
        raise RuntimeError(
            "Unrecognised %s type (%s)" % (self.__class__.__name__.lower(), self.type)
        )

    def __str__(self):
        return self.name

    @property
    def name(self):
        """
        A text description of the benefit/condition. Every proxy class
        has to implement it.

        This is used in the dropdowns within the offer dashboard.
        """
        proxy_instance = self.proxy()
        if self.proxy_class and self.__class__ == proxy_instance.__class__:
            raise AssertionError("Name property is not defined on proxy class.")
        return proxy_instance.name

    @property
    def description(self):
        """
        A description of the benefit/condition.
        Defaults to the name. May contain HTML.
        """
        try:
            return self.name
        except AttributeError:
            pass


class AbstractConditionalOffer(models.Model):
    """
    A conditional offer (e.g. buy 1, get 10% off)
    """

    name = models.CharField(
        _("Name"),
        max_length=128,
        unique=True,
        help_text=_("This is displayed within the customer's basket"),
    )
    slug = fields.AutoSlugField(
        _("Slug"), max_length=128, unique=True, populate_from="name"
    )
    description = models.TextField(
        _("Description"),
        blank=True,
        help_text=_("This is displayed on the offer browsing page"),
    )

    # Offers come in a few different types:
    # (a) Offers that are available to all customers on the site. e.g. a
    #     3-for-2 offer.
    # (b) Offers that are linked to a voucher, and only become available once
    #     that voucher has been applied to the basket
    # (c) Offers that are linked to a user.  e.g. all students get 10% off.  The
    #     code to apply this offer needs to be coded
    # (d) Session offers - these are temporarily available to a user after some
    #     trigger event.  e.g. users coming from some affiliate site get 10%
    #     off.
    SITE, VOUCHER, USER, SESSION = ("Site", "Voucher", "User", "Session")
    TYPE_CHOICES = (
        (SITE, _("Site offer - available to all users")),
        (
            VOUCHER,
            _(
                "Voucher offer - only available after entering "
                "the appropriate voucher code"
            ),
        ),
        (USER, _("User offer - available to certain types of user")),
        (
            SESSION,
            _(
                "Session offer - temporary offer, available for "
                "a user for the duration of their session"
            ),
        ),
    )
    offer_type = models.CharField(
        _("Type"), choices=TYPE_CHOICES, default=SITE, max_length=128
    )

    exclusive = models.BooleanField(
        _("Exclusive offer"),
        help_text=_("Exclusive offers cannot be combined on the same items"),
        default=True,
    )
    combinations = models.ManyToManyField(
        "offer.ConditionalOffer",
        help_text=_(
            "Select other non-exclusive offers that this offer can be combined with on the same items"
        ),
        related_name="in_combination",
        limit_choices_to={"exclusive": False},
        blank=True,
    )

    # We track a status variable so it's easier to load offers that are
    # 'available' in some sense.
    OPEN, SUSPENDED, CONSUMED = "Open", "Suspended", "Consumed"
    status = models.CharField(_("Status"), max_length=64, default=OPEN)

    condition = models.ForeignKey(
        "offer.Condition",
        on_delete=models.CASCADE,
        related_name="offers",
        verbose_name=_("Condition"),
    )
    benefit = models.ForeignKey(
        "offer.Benefit",
        on_delete=models.CASCADE,
        related_name="offers",
        verbose_name=_("Benefit"),
    )

    # Some complicated situations require offers to be applied in a set order.
    priority = models.IntegerField(
        _("Priority"),
        default=0,
        db_index=True,
        help_text=_("The highest priority offers are applied first"),
    )

    # AVAILABILITY

    # Range of availability.  Note that if this is a voucher offer, then these
    # dates are ignored and only the dates from the voucher are used to
    # determine availability.
    start_datetime = models.DateTimeField(
        _("Start date"),
        blank=True,
        null=True,
        help_text=_(
            "Offers are active from the start date. "
            "Leave this empty if the offer has no start date."
        ),
    )
    end_datetime = models.DateTimeField(
        _("End date"),
        blank=True,
        null=True,
        help_text=_(
            "Offers are active until the end date. "
            "Leave this empty if the offer has no expiry date."
        ),
    )

    # Use this field to limit the number of times this offer can be applied in
    # total.  Note that a single order can apply an offer multiple times so
    # this is not necessarily the same as the number of orders that can use it.
    # Also see max_basket_applications.
    max_global_applications = models.PositiveIntegerField(
        _("Max global applications"),
        help_text=_(
            "The number of times this offer can be used before it is unavailable"
        ),
        blank=True,
        null=True,
    )

    # Use this field to limit the number of times this offer can be used by a
    # single user.  This only works for signed-in users - it doesn't really
    # make sense for sites that allow anonymous checkout.
    max_user_applications = models.PositiveIntegerField(
        _("Max user applications"),
        help_text=_("The number of times a single user can use this offer"),
        blank=True,
        null=True,
    )

    # Use this field to limit the number of times this offer can be applied to
    # a basket (and hence a single order). Often, an offer should only be
    # usable once per basket/order, so this field will commonly be set to 1.
    max_basket_applications = models.PositiveIntegerField(
        _("Max basket applications"),
        blank=True,
        null=True,
        help_text=_(
            "The number of times this offer can be applied to a basket (and order)"
        ),
    )

    # Use this field to limit the amount of discount an offer can lead to.
    # This can be helpful with budgeting.
    max_discount = models.DecimalField(
        _("Max discount"),
        decimal_places=2,
        max_digits=12,
        null=True,
        blank=True,
        help_text=_(
            "When an offer has given more discount to orders "
            "than this threshold, then the offer becomes "
            "unavailable"
        ),
    )

    # TRACKING
    # These fields are used to enforce the limits set by the
    # max_* fields above.

    total_discount = models.DecimalField(
        _("Total Discount"), decimal_places=2, max_digits=12, default=D("0.00")
    )
    num_applications = models.PositiveIntegerField(
        _("Number of applications"), default=0
    )
    num_orders = models.PositiveIntegerField(_("Number of Orders"), default=0)

    redirect_url = fields.ExtendedURLField(_("URL redirect (optional)"), blank=True)
    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)

    objects = models.Manager()
    active = ActiveOfferManager()

    # We need to track the voucher that this offer came from (if it is a
    # voucher offer)
    _voucher = None

    class Meta:
        abstract = True
        app_label = "offer"
        ordering = ["-priority", "pk"]
        verbose_name = _("Conditional offer")
        verbose_name_plural = _("Conditional offers")

    def save(self, *args, **kwargs):
        # Check to see if consumption thresholds have been broken
        if not self.is_suspended:
            if self.get_max_applications() == 0:
                self.status = self.CONSUMED
            else:
                self.status = self.OPEN

        return super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse("offer:detail", kwargs={"slug": self.slug})

    def __str__(self):
        return self.name

    def clean(self):
        if (
            self.start_datetime
            and self.end_datetime
            and self.start_datetime > self.end_datetime
        ):
            raise exceptions.ValidationError(
                _("End date should be later than start date")
            )

    @property
    def is_voucher_offer_type(self):
        return self.offer_type == self.VOUCHER

    @property
    def is_open(self):
        return self.status == self.OPEN

    @property
    def is_suspended(self):
        return self.status == self.SUSPENDED

    def suspend(self):
        self.status = self.SUSPENDED
        self.save()

    suspend.alters_data = True

    def unsuspend(self):
        self.status = self.OPEN
        self.save()

    unsuspend.alters_data = True

    def is_available(self, user=None, test_date=None):
        """
        Test whether this offer is available to be used
        """
        if self.is_suspended:
            return False
        if test_date is None:
            test_date = now()
        predicates = []
        if self.start_datetime:
            predicates.append(self.start_datetime > test_date)
        if self.end_datetime:
            predicates.append(test_date > self.end_datetime)
        if any(predicates):
            return False
        return self.get_max_applications(user) > 0

    def is_condition_satisfied(self, basket):
        return self.condition.proxy().is_satisfied(self, basket)

    def is_condition_partially_satisfied(self, basket):
        return self.condition.proxy().is_partially_satisfied(self, basket)

    def get_upsell_message(self, basket):
        return self.condition.proxy().get_upsell_message(self, basket)

    def apply_benefit(self, basket):
        """
        Applies the benefit to the given basket and returns the discount.
        """
        if not self.is_condition_satisfied(basket):
            return ZERO_DISCOUNT
        return self.benefit.proxy().apply(basket, self.condition.proxy(), self)

    def apply_deferred_benefit(self, basket, order, application):
        """
        Applies any deferred benefits.  These are things like adding loyalty
        points to someone's account.
        """
        return self.benefit.proxy().apply_deferred(basket, order, application)

    def set_voucher(self, voucher):
        self._voucher = voucher

    def get_voucher(self):
        return self._voucher

    def get_max_applications(self, user=None):
        """
        Return the number of times this offer can be applied to a basket for a
        given user.
        """
        if self.max_discount and self.total_discount >= self.max_discount:
            return 0

        # Hard-code a maximum value as we need some sensible upper limit for
        # when there are not other caps.
        limits = [10000]
        if self.max_user_applications and user:
            limits.append(
                max(
                    0, self.max_user_applications - self.get_num_user_applications(user)
                )
            )
        if self.max_basket_applications:
            limits.append(self.max_basket_applications)
        if self.max_global_applications:
            limits.append(max(0, self.max_global_applications - self.num_applications))
        return min(limits)

    def get_num_user_applications(self, user):
        OrderDiscount = get_model("order", "OrderDiscount")
        aggregates = OrderDiscount.objects.filter(
            offer_id=self.id, order__user=user
        ).aggregate(total=models.Sum("frequency"))
        return aggregates["total"] if aggregates["total"] is not None else 0

    def shipping_discount(self, charge, currency=None):
        return self.benefit.proxy().shipping_discount(charge, currency)

    def record_usage(self, discount):
        self.num_applications += discount["freq"]
        self.total_discount += discount["discount"]
        self.num_orders += 1
        self.save()

    record_usage.alters_data = True

    def availability_description(self):
        """
        Return a description of when this offer is available
        """
        restrictions = self.availability_restrictions()
        descriptions = [r["description"] for r in restrictions]
        return "<br/>".join(descriptions)

    def availability_restrictions(self):
        restrictions = []
        if self.is_suspended:
            restrictions.append(
                {"description": _("Offer is suspended"), "is_satisfied": False}
            )

        if self.max_global_applications:
            remaining = self.max_global_applications - self.num_applications
            desc = _("Limited to %(total)d uses (%(remainder)d remaining)") % {
                "total": self.max_global_applications,
                "remainder": remaining,
            }
            restrictions.append({"description": desc, "is_satisfied": remaining > 0})

        if self.max_user_applications:
            if self.max_user_applications == 1:
                desc = _("Limited to 1 use per user")
            else:
                desc = _("Limited to %(total)d uses per user") % {
                    "total": self.max_user_applications
                }
            restrictions.append({"description": desc, "is_satisfied": True})

        if self.max_basket_applications:
            if self.max_user_applications == 1:
                desc = _("Limited to 1 use per basket")
            else:
                desc = _("Limited to %(total)d uses per basket") % {
                    "total": self.max_basket_applications
                }
            restrictions.append({"description": desc, "is_satisfied": True})

        def hide_time_if_zero(dt):
            # Only show hours/minutes if they have been specified
            if dt.tzinfo:
                localtime = dt.astimezone(get_current_timezone())
            else:
                localtime = dt
            if localtime.hour == 0 and localtime.minute == 0:
                return date_filter(localtime, settings.DATE_FORMAT)
            return date_filter(localtime, settings.DATETIME_FORMAT)

        if self.start_datetime or self.end_datetime:
            today = now()
            is_satisfied = False
            if self.start_datetime and self.end_datetime:
                desc = _("Available between %(start)s and %(end)s") % {
                    "start": hide_time_if_zero(self.start_datetime),
                    "end": hide_time_if_zero(self.end_datetime),
                }
                is_satisfied = self.start_datetime <= today <= self.end_datetime
            elif self.start_datetime:
                desc = _("Available from %(start)s") % {
                    "start": hide_time_if_zero(self.start_datetime)
                }
                is_satisfied = today >= self.start_datetime
            elif self.end_datetime:
                desc = _("Available until %(end)s") % {
                    "end": hide_time_if_zero(self.end_datetime)
                }
                is_satisfied = today <= self.end_datetime
            restrictions.append({"description": desc, "is_satisfied": is_satisfied})

        if self.max_discount:
            desc = _("Limited to a cost of %(max)s") % {
                "max": currency(self.max_discount)
            }
            restrictions.append(
                {
                    "description": desc,
                    "is_satisfied": self.total_discount < self.max_discount,
                }
            )

        return restrictions

    @property
    def has_products(self):
        return self.condition.range is not None

    def products(self):
        """
        Return a queryset of products in this offer
        """
        Product = get_model("catalogue", "Product")
        if not self.has_products:
            return Product.objects.none()

        queryset = self.condition.range.all_products()
        return queryset.filter(is_discountable=True).browsable()

    @cached_property
    def combined_offers(self):
        return self.__class__.objects.filter(
            models.Q(pk=self.pk)
            | models.Q(pk__in=self.combinations.values_list("pk", flat=True))
            | models.Q(pk__in=self.in_combination.values_list("pk", flat=True))
        ).distinct()


class AbstractBenefit(BaseOfferMixin, models.Model):
    range = models.ForeignKey(
        "offer.Range",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("Range"),
    )

    # Benefit types
    PERCENTAGE, FIXED, FIXED_UNIT, MULTIBUY, FIXED_PRICE = (
        "Percentage",
        "Absolute",
        "Fixed",
        "Multibuy",
        "Fixed price",
    )
    SHIPPING_PERCENTAGE, SHIPPING_ABSOLUTE, SHIPPING_FIXED_PRICE = (
        "Shipping percentage",
        "Shipping absolute",
        "Shipping fixed price",
    )
    TYPE_CHOICES = (
        (PERCENTAGE, _("Discount is a percentage off of the product's value")),
        (FIXED, _("Discount is a fixed amount off of the basket's total")),
        (FIXED_UNIT, _("Discount is a fixed amount off of the product's value")),
        (MULTIBUY, _("Discount is to give the cheapest product for free")),
        (FIXED_PRICE, _("Get the products that meet the condition for a fixed price")),
        (SHIPPING_ABSOLUTE, _("Discount is a fixed amount of the shipping cost")),
        (SHIPPING_FIXED_PRICE, _("Get shipping for a fixed price")),
        (
            SHIPPING_PERCENTAGE,
            _("Discount is a percentage off of the shipping cost"),
        ),
    )
    type = models.CharField(_("Type"), max_length=128, choices=TYPE_CHOICES, blank=True)

    # The value to use with the designated type.  This can be either an integer
    # (eg for multibuy) or a decimal (eg an amount) which is slightly
    # confusing.
    value = fields.PositiveDecimalField(
        _("Value"), decimal_places=2, max_digits=12, null=True, blank=True
    )

    # If this is not set, then there is no upper limit on how many products
    # can be discounted by this benefit.
    max_affected_items = models.PositiveIntegerField(
        _("Max Affected Items"),
        blank=True,
        null=True,
        help_text=_(
            "Set this to prevent the discount consuming all items "
            "within the range that are in the basket."
        ),
    )

    # A custom benefit class can be used instead.  This means the
    # type/value/max_affected_items fields should all be None.
    proxy_class = fields.NullCharField(_("Custom class"), max_length=255, default=None)

    class Meta:
        abstract = True
        app_label = "offer"
        verbose_name = _("Benefit")
        verbose_name_plural = _("Benefits")

    @property
    def proxy_map(self):
        return {
            self.PERCENTAGE: get_class("offer.benefits", "PercentageDiscountBenefit"),
            self.FIXED: get_class("offer.benefits", "AbsoluteDiscountBenefit"),
            self.FIXED_UNIT: get_class("offer.benefits", "FixedUnitDiscountBenefit"),
            self.MULTIBUY: get_class("offer.benefits", "MultibuyDiscountBenefit"),
            self.FIXED_PRICE: get_class("offer.benefits", "FixedPriceBenefit"),
            self.SHIPPING_ABSOLUTE: get_class(
                "offer.benefits", "ShippingAbsoluteDiscountBenefit"
            ),
            self.SHIPPING_FIXED_PRICE: get_class(
                "offer.benefits", "ShippingFixedPriceBenefit"
            ),
            self.SHIPPING_PERCENTAGE: get_class(
                "offer.benefits", "ShippingPercentageDiscountBenefit"
            ),
        }

    def apply(self, basket, condition, offer):
        return ZERO_DISCOUNT

    def apply_deferred(self, basket, order, application):
        return None

    def clean(self):
        if not self.type:
            return
        method_name = "clean_%s" % self.type.lower().replace(" ", "_")
        if hasattr(self, method_name):
            getattr(self, method_name)()

    def clean_multibuy(self):
        errors = []

        if not self.range:
            errors.append(_("Multibuy benefits require a product range"))
        if self.value:
            errors.append(_("Multibuy benefits don't require a value"))
        if self.max_affected_items:
            errors.append(
                _("Multibuy benefits don't require a 'max affected items' attribute")
            )

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_percentage(self):
        errors = []

        if not self.range:
            errors.append(_("Percentage benefits require a product range"))

        if not self.value:
            errors.append(_("Percentage discount benefits require a value"))
        elif self.value > 100:
            errors.append(_("Percentage discount cannot be greater than 100"))

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_shipping_absolute(self):
        errors = []
        if not self.value:
            errors.append(_("A discount value is required"))
        if self.range:
            errors.append(
                _(
                    "No range should be selected as this benefit does "
                    "not apply to products"
                )
            )
        if self.max_affected_items:
            errors.append(
                _(
                    "Shipping discounts don't require a "
                    "'max affected items' attribute"
                )
            )

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_shipping_percentage(self):
        errors = []

        if not self.value:
            errors.append(_("Percentage discount benefits require a value"))
        elif self.value > 100:
            errors.append(_("Percentage discount cannot be greater than 100"))

        if self.range:
            errors.append(
                _(
                    "No range should be selected as this benefit does "
                    "not apply to products"
                )
            )
        if self.max_affected_items:
            errors.append(
                _(
                    "Shipping discounts don't require a "
                    "'max affected items' attribute"
                )
            )
        if errors:
            raise exceptions.ValidationError(errors)

    def clean_shipping_fixed_price(self):
        errors = []
        if self.range:
            errors.append(
                _(
                    "No range should be selected as this benefit does "
                    "not apply to products"
                )
            )
        if self.max_affected_items:
            errors.append(
                _(
                    "Shipping discounts don't require a "
                    "'max affected items' attribute"
                )
            )

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_fixed_price(self):
        if self.range:
            raise exceptions.ValidationError(
                _(
                    "No range should be selected as the condition range will "
                    "be used instead."
                )
            )

    def clean_absolute(self):
        errors = []
        if not self.range:
            errors.append(_("Fixed discount benefits require a product range"))
        if not self.value:
            errors.append(_("Fixed discount benefits require a value"))

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_fixed(self):
        errors = []
        if not self.range:
            errors.append(
                _("Fixed product level discount benefits require a product range")
            )
        if not self.value:
            errors.append(_("Fixed product level discount benefits require a value"))

        if errors:
            raise exceptions.ValidationError(errors)

    def round(self, amount, currency=None, round_type=ROUND_DOWN):
        """
        Apply rounding to discount amount
        """
        rounding_function_path = getattr(
            settings, "OSCAR_OFFER_ROUNDING_FUNCTION", None
        )
        if rounding_function_path:
            rounding_function = cached_import_string(rounding_function_path)
            return rounding_function(amount, currency)

        return amount.quantize(D(".01"), round_type)

    def _effective_max_affected_items(self):
        """
        Return the maximum number of items that can have a discount applied
        during the application of this benefit
        """
        return self.max_affected_items if self.max_affected_items else 10000

    def can_apply_benefit(self, line):
        """
        Determines whether the benefit can be applied to a given basket line
        """
        return line.stockrecord and line.product.is_discountable

    # pylint: disable=W0622
    def get_applicable_lines(self, offer, basket, range=None):
        """
        Return the basket lines that are available to be discounted

        :basket: The basket
        :range: The range of products to use for filtering.  The fixed-price
                benefit ignores its range and uses the condition range
        """
        if range is None:
            range = self.range
        line_tuples = []
        for line in basket.all_lines():
            product = line.product

            if not range.contains_product(product) or not self.can_apply_benefit(line):
                continue

            price = unit_price(offer, line)
            if not price:
                # Avoid zero price products
                continue
            line_tuples.append((price, line))

        # We sort lines to be cheapest first to ensure consistent applications
        return sorted(line_tuples, key=operator.itemgetter(0))

    def shipping_discount(self, charge, currency=None):
        return D("0.00")


class AbstractCondition(BaseOfferMixin, models.Model):
    """
    A condition for an offer to be applied. You can either specify a custom
    proxy class, or need to specify a type, range and value.
    """

    COUNT, VALUE, COVERAGE = ("Count", "Value", "Coverage")
    TYPE_CHOICES = (
        (
            COUNT,
            _("Depends on number of items in basket that are in condition range"),
        ),
        (
            VALUE,
            _("Depends on value of items in basket that are in condition range"),
        ),
        (
            COVERAGE,
            _(
                "Needs to contain a set number of DISTINCT items "
                "from the condition range"
            ),
        ),
    )
    range = models.ForeignKey(
        "offer.Range",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("Range"),
    )
    type = models.CharField(_("Type"), max_length=128, choices=TYPE_CHOICES, blank=True)
    value = fields.PositiveDecimalField(
        _("Value"), decimal_places=2, max_digits=12, null=True, blank=True
    )

    proxy_class = fields.NullCharField(_("Custom class"), max_length=255, default=None)

    class Meta:
        abstract = True
        app_label = "offer"
        verbose_name = _("Condition")
        verbose_name_plural = _("Conditions")

    @property
    def proxy_map(self):
        return {
            self.COUNT: get_class("offer.conditions", "CountCondition"),
            self.VALUE: get_class("offer.conditions", "ValueCondition"),
            self.COVERAGE: get_class("offer.conditions", "CoverageCondition"),
        }

    def clean(self):
        # The form will validate whether this is ok or not.
        if not self.type:
            return
        method_name = "clean_%s" % self.type.lower()
        if hasattr(self, method_name):
            getattr(self, method_name)()

    def clean_count(self):
        errors = []

        if not self.range:
            errors.append(_("Count conditions require a product range"))

        if not self.value:
            errors.append(_("Count conditions require a value"))

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_value(self):
        errors = []

        if not self.range:
            errors.append(_("Value conditions require a product range"))

        if not self.value:
            errors.append(_("Value conditions require a value"))

        if errors:
            raise exceptions.ValidationError(errors)

    def clean_coverage(self):
        errors = []

        if not self.range:
            errors.append(_("Coverage conditions require a product range"))

        if not self.value:
            errors.append(_("Coverage conditions require a value"))

        if errors:
            raise exceptions.ValidationError(errors)

    def consume_items(self, offer, basket, affected_lines):
        pass

    def is_satisfied(self, offer, basket):
        """
        Determines whether a given basket meets this condition.  This is
        stubbed in this top-class object.  The subclassing proxies are
        responsible for implementing it correctly.
        """
        return False

    def is_partially_satisfied(self, offer, basket):
        """
        Determine if the basket partially meets the condition.  This is useful
        for up-selling messages to entice customers to buy something more in
        order to qualify for an offer.
        """
        return False

    def get_upsell_message(self, offer, basket):
        return None

    def can_apply_condition(self, line):
        """
        Determines whether the condition can be applied to a given basket line
        """
        if not line.stockrecord_id:
            return False
        product = line.product
        return self.range.contains_product(product) and product.is_discountable

    def get_applicable_lines(self, offer, basket, most_expensive_first=True):
        """
        Return line data for the lines that can be consumed by this condition
        """
        line_tuples = []
        for line in basket.all_lines():
            if not self.can_apply_condition(line):
                continue

            price = unit_price(offer, line)
            if not price:
                continue
            line_tuples.append((price, line))
        key = operator.itemgetter(0)
        if most_expensive_first:
            return sorted(line_tuples, reverse=True, key=key)
        return sorted(line_tuples, key=key)


class AbstractRange(models.Model):
    """
    Represents a range of products that can be used within an offer.
    """

    name = models.CharField(_("Name"), max_length=128, unique=True)
    slug = fields.AutoSlugField(
        _("Slug"), max_length=128, unique=True, populate_from="name"
    )

    description = models.TextField(_("Description"), blank=True)

    # Whether this range is public
    is_public = models.BooleanField(
        _("Is public?"),
        default=False,
        help_text=_("Public ranges have a customer-facing page"),
    )

    includes_all_products = models.BooleanField(
        _("Includes all products?"), default=False
    )

    included_products = models.ManyToManyField(
        "catalogue.Product",
        related_name="includes",
        blank=True,
        verbose_name=_("Included Products"),
        through="offer.RangeProduct",
    )
    excluded_products = models.ManyToManyField(
        "catalogue.Product",
        related_name="excludes",
        blank=True,
        verbose_name=_("Excluded Products"),
    )
    classes = models.ManyToManyField(
        "catalogue.ProductClass",
        related_name="classes",
        blank=True,
        verbose_name=_("Product Types"),
    )
    included_categories = models.ManyToManyField(
        "catalogue.Category",
        related_name="includes",
        blank=True,
        verbose_name=_("Included Categories"),
    )
    excluded_categories = models.ManyToManyField(
        "catalogue.Category",
        related_name="excludes",
        blank=True,
        verbose_name=_("Excluded Categories"),
        help_text=_(
            "Products with these categories are excluded from the range when "
            "Includes all products is checked"
        ),
    )

    # Allow a custom range instance to be specified
    proxy_class = fields.NullCharField(
        _("Custom class"), max_length=255, default=None, unique=True
    )

    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)

    objects = RangeManager()
    browsable = BrowsableRangeManager()

    class Meta:
        abstract = True
        app_label = "offer"
        ordering = ["name"]
        verbose_name = _("Range")
        verbose_name_plural = _("Ranges")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("catalogue:range", kwargs={"slug": self.slug})

    @cached_property
    def proxy(self):
        if self.proxy_class:
            return load_proxy(self.proxy_class)()

    def add_product(self, product, display_order=None):
        """Add product to the range

        When adding product that is already in the range, prevent re-adding it.
        If display_order is specified, update it.

        Default display_order for a new product in the range is 0; this puts
        the product at the top of the list.
        """

        initial_order = display_order or 0
        RangeProduct = self.included_products.through
        relation, __ = RangeProduct.objects.get_or_create(
            range=self, product=product, defaults={"display_order": initial_order}
        )

        if display_order is not None and relation.display_order != display_order:
            relation.display_order = display_order
            relation.save()

        # Remove product from excluded products if it was removed earlier and
        # re-added again, thus it returns back to the range product list.
        self.excluded_products.remove(product)

        # invalidate cache because queryset has changed
        self.invalidate_cached_queryset()

    def remove_product(self, product):
        """
        Remove product from range. To save on queries, this function does not
        check if the product is in fact in the range.
        """
        RangeProduct = self.included_products.through
        RangeProduct.objects.filter(range=self, product=product).delete()
        # Making sure product will be excluded from range products list by adding to
        # respective field. Otherwise, it could be included as a product from included
        # category or etc.
        self.excluded_products.add(product)

        # invalidate cache because queryset has changed
        self.invalidate_cached_queryset()

    def contains_product(self, product):
        if self.proxy:
            return self.proxy.contains_product(product)
        return self.product_queryset.filter(id=product.id).exists()

    def invalidate_cached_queryset(self):
        try:
            del self.product_queryset
        except AttributeError:
            pass

    def num_products(self):
        # Delegate to a proxy class if one is provided
        if self.proxy:
            return self.proxy.num_products()
        if self.includes_all_products:
            return None
        return self.all_products().count()

    def all_products(self):
        """
        Return a queryset containing all the products in the range

        This includes included_products plus the products contained in the
        included classes and categories, minus the products in
        excluded_products.
        """
        if self.proxy:
            return self.proxy.all_products()

        return self.product_queryset

    @cached_property
    def product_queryset(self):
        "cached queryset of all the products in the Range"
        ProductCategoryHierarchy = get_model("catalogue", "ProductCategoryHierarchy")

        Product = self.included_products.model

        if self.includes_all_products:
            _filter = Q(id__in=self.excluded_products.values("id"))
            # extend filter if excluded_categories exist
            if self.excluded_categories.exists():
                if use_productcategory_materialised_view():
                    product_ids = ProductCategoryHierarchy.objects.filter(
                        category_id__in=self.excluded_categories.values_list(
                            "id", flat=True
                        )
                    ).values_list("product_id", flat=True)

                    _filter |= Q(id__in=product_ids)
                    _filter |= Q(parent__id__in=product_ids)
                else:
                    expanded_range_categories = ExpandDownwardsCategoryQueryset(
                        self.excluded_categories.values("id")
                    )
                    _filter |= Q(categories__in=expanded_range_categories)
                    # extend filter for parent categories, exclude parent = None
                    if (
                        Product.objects.exclude(parent=None)
                        .filter(parent__categories__in=expanded_range_categories)
                        .exists()
                    ):
                        _filter |= Q(parent__categories__in=expanded_range_categories)

            # Filter out blacklisted
            return Product.objects.exclude(_filter)

        # start with filter clause that always applies
        _filter = Q(includes=self)

        # extend filter if included_products have children
        if Product.objects.filter(parent__includes=self).exists():
            _filter |= Q(parent__includes=self)

        # extend filter if included classes exist:
        if self.classes.exists():
            _filter |= Q(product_class__classes=self)
            # this is always very fast so no check is needed
            _filter |= Q(parent__product_class__classes=self)

        # extend filter if included_categories exist
        if self.included_categories.exists():
            if use_productcategory_materialised_view():
                product_ids = ProductCategoryHierarchy.objects.filter(
                    category_id__in=self.included_categories.values_list(
                        "id", flat=True
                    )
                ).values_list("product_id", flat=True)

                _filter |= Q(id__in=product_ids)
                _filter |= Q(parent__id__in=product_ids)

            else:
                expanded_range_categories = ExpandDownwardsCategoryQueryset(
                    self.included_categories.values("id")
                )
                _filter |= Q(categories__in=expanded_range_categories)
                # extend filter for parent categories, exclude parent = None
                if (
                    Product.objects.exclude(parent=None)
                    .filter(parent__categories__in=expanded_range_categories)
                    .exists()
                ):
                    _filter |= Q(parent__categories__in=expanded_range_categories)

        qs = Product.objects.filter(_filter, ~Q(excludes=self))

        if Product.objects.filter(parent__excludes=self).exists():
            qs = qs.filter(~Q(parent__excludes=self))

        # make sure to filter out duplicates originating from a join
        return qs.distinct()

    @property
    def is_editable(self):
        """
        Test whether this range can be edited in the dashboard.
        """
        return not self.proxy_class

    @property
    def is_reorderable(self):
        """
        Test whether products for the range can be re-ordered.
        """
        return not (
            self.included_categories.exists()
            or self.classes.exists()
            or (self.excluded_categories.exists() and self.includes_all_products)
        )


class AbstractRangeProduct(models.Model):
    """
    Allow ordering products inside ranges
    Exists to allow customising.
    """

    range = models.ForeignKey("offer.Range", on_delete=models.CASCADE)
    product = models.ForeignKey("catalogue.Product", on_delete=models.CASCADE)
    display_order = models.IntegerField(default=0)

    class Meta:
        abstract = True
        app_label = "offer"
        unique_together = ("range", "product")


class AbstractRangeProductFileUpload(models.Model):
    range = models.ForeignKey(
        "offer.Range",
        on_delete=models.CASCADE,
        related_name="file_uploads",
        verbose_name=_("Range"),
    )
    filepath = models.CharField(_("File Path"), max_length=255)
    size = models.PositiveIntegerField(_("Size"))
    uploaded_by = models.ForeignKey(
        AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("Uploaded By")
    )
    date_uploaded = models.DateTimeField(
        _("Date Uploaded"), auto_now_add=True, db_index=True
    )

    INCLUDED_PRODUCTS_TYPE = "included"
    EXCLUDED_PRODUCTS_TYPE = "excluded"
    UPLOAD_TYPE_CHOICES = [
        (INCLUDED_PRODUCTS_TYPE, "Included products upload"),
        (EXCLUDED_PRODUCTS_TYPE, "Excluded products upload"),
    ]
    upload_type = models.CharField(
        max_length=8, choices=UPLOAD_TYPE_CHOICES, default=INCLUDED_PRODUCTS_TYPE
    )

    PENDING, FAILED, PROCESSED = "Pending", "Failed", "Processed"
    choices = (
        (PENDING, PENDING),
        (FAILED, FAILED),
        (PROCESSED, PROCESSED),
    )
    status = models.CharField(
        _("Status"), max_length=32, choices=choices, default=PENDING
    )
    error_message = models.CharField(_("Error Message"), max_length=255, blank=True)

    # Post-processing audit fields
    date_processed = models.DateTimeField(_("Date Processed"), null=True)
    num_new_skus = models.PositiveIntegerField(_("Number of New SKUs"), null=True)
    num_unknown_skus = models.PositiveIntegerField(
        _("Number of Unknown SKUs"), null=True
    )
    num_duplicate_skus = models.PositiveIntegerField(
        _("Number of Duplicate SKUs"), null=True
    )

    class Meta:
        abstract = True
        app_label = "offer"
        ordering = ("-date_uploaded",)
        verbose_name = _("Range Product Uploaded File")
        verbose_name_plural = _("Range Product Uploaded Files")

    def mark_as_failed(self, message=None):
        self.date_processed = now()
        self.error_message = message
        self.status = self.FAILED
        self.save()

    def mark_as_processed(self, num_new, num_unknown, num_duplicate):
        self.status = self.PROCESSED
        self.date_processed = now()
        self.num_new_skus = num_new
        self.num_unknown_skus = num_unknown
        self.num_duplicate_skus = num_duplicate
        self.save()

    def was_processing_successful(self):
        return self.status == self.PROCESSED

    def process(self, file_obj):
        """
        Process the file upload and add products to the range
        or add products to range.excluded_products
        """
        all_ids = set(self.extract_ids(file_obj))
        if self.upload_type == self.INCLUDED_PRODUCTS_TYPE:
            products = self.range.all_products()
        elif self.upload_type == self.EXCLUDED_PRODUCTS_TYPE:
            products = self.range.excluded_products.all()
        else:
            raise ValueError("Unable to process upload type: %s" % self.upload_type)
        existing_skus = products.values_list("stockrecords__partner_sku", flat=True)
        existing_skus = set(filter(bool, existing_skus))
        existing_upcs = products.values_list("upc", flat=True)
        existing_upcs = set(filter(bool, existing_upcs))
        existing_ids = existing_skus.union(existing_upcs)
        new_ids = all_ids - existing_ids

        Product = get_model("catalogue", "Product")
        products = Product._default_manager.filter(
            models.Q(stockrecords__partner_sku__in=new_ids) | models.Q(upc__in=new_ids)
        )
        for product in products:
            if self.upload_type == self.INCLUDED_PRODUCTS_TYPE:
                self.range.add_product(product)
            else:
                self.range.excluded_products.add(product)

        # Processing stats
        found_skus = products.values_list("stockrecords__partner_sku", flat=True)
        found_skus = set(filter(bool, found_skus))
        found_upcs = set(filter(bool, products.values_list("upc", flat=True)))
        found_ids = found_skus.union(found_upcs)
        missing_ids = new_ids - found_ids
        dupes = set(all_ids).intersection(existing_ids)

        self.mark_as_processed(products.count(), len(missing_ids), len(dupes))
        return products

    def extract_ids(self, file_obj):
        reader = csv.reader(file_obj)
        for line in reader:
            if line:
                yield from line



from django.contrib import admin

from oscar.core.loading import get_model

WishList = get_model("wishlists", "WishList")
Line = get_model("wishlists", "Line")
WishListSharedEmail = get_model("wishlists", "WishListSharedEmail")


admin.site.register(WishList)
admin.site.register(Line)
admin.site.register(WishListSharedEmail)



from django.conf import settings
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy
from django_tables2 import A, Column, LinkColumn, TemplateColumn

from oscar.core.loading import get_class, get_model

DashboardTable = get_class("dashboard.tables", "DashboardTable")
Product = get_model("catalogue", "Product")
Category = get_model("catalogue", "Category")
AttributeOptionGroup = get_model("catalogue", "AttributeOptionGroup")
Option = get_model("catalogue", "Option")


class ProductTable(DashboardTable):
    title = TemplateColumn(
        verbose_name=_("Title"),
        template_name="oscar/dashboard/catalogue/product_row_title.html",
        order_by="title",
        accessor=A("title"),
    )
    image = TemplateColumn(
        verbose_name=_("Image"),
        template_name="oscar/dashboard/catalogue/product_row_image.html",
        orderable=False,
    )
    product_class = Column(
        verbose_name=_("Product type"),
        accessor=A("product_class"),
        order_by="product_class__name",
    )
    variants = TemplateColumn(
        verbose_name=_("Variants"),
        template_name="oscar/dashboard/catalogue/product_row_variants.html",
        orderable=False,
    )
    stock_records = TemplateColumn(
        verbose_name=_("Stock records"),
        template_name="oscar/dashboard/catalogue/product_row_stockrecords.html",
        orderable=False,
    )
    actions = TemplateColumn(
        verbose_name=_("Actions"),
        template_name="oscar/dashboard/catalogue/product_row_actions.html",
        orderable=False,
    )

    icon = "fas fa-sitemap"

    class Meta(DashboardTable.Meta):
        model = Product
        fields = ("upc", "is_public", "date_updated")
        sequence = (
            "title",
            "upc",
            "image",
            "product_class",
            "variants",
            "stock_records",
            "...",
            "is_public",
            "date_updated",
            "actions",
        )
        order_by = "-date_updated"


class CategoryTable(DashboardTable):
    name = LinkColumn("dashboard:catalogue-category-update", args=[A("pk")])
    description = TemplateColumn(
        template_code='{{ record.description|default:""|striptags'
        '|cut:"&nbsp;"|truncatewords:6 }}'
    )
    # mark_safe is needed because of
    # https://github.com/bradleyayers/django-tables2/issues/187
    num_children = LinkColumn(
        "dashboard:catalogue-category-detail-list",
        args=[A("pk")],
        verbose_name=mark_safe(_("Number of child categories")),
        accessor="get_num_children",
        orderable=False,
    )
    actions = TemplateColumn(
        template_name="oscar/dashboard/catalogue/category_row_actions.html",
        orderable=False,
    )

    icon = "sitemap"
    caption = ngettext_lazy("%s Category", "%s Categories")

    class Meta(DashboardTable.Meta):
        model = Category
        fields = ("name", "description", "is_public")
        sequence = ("name", "description", "...", "is_public", "actions")


class AttributeOptionGroupTable(DashboardTable):
    name = TemplateColumn(
        verbose_name=_("Name"),
        template_name="oscar/dashboard/catalogue/attribute_option_group_row_name.html",
        order_by="name",
    )
    option_summary = TemplateColumn(
        verbose_name=_("Option summary"),
        template_name="oscar/dashboard/catalogue/attribute_option_group_row_option_summary.html",
        orderable=False,
    )
    actions = TemplateColumn(
        verbose_name=_("Actions"),
        template_name="oscar/dashboard/catalogue/attribute_option_group_row_actions.html",
        orderable=False,
    )

    icon = "sitemap"
    caption = ngettext_lazy("%s Attribute Option Group", "%s Attribute Option Groups")

    class Meta(DashboardTable.Meta):
        model = AttributeOptionGroup
        fields = ("name",)
        sequence = ("name", "option_summary", "actions")
        per_page = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE


class OptionTable(DashboardTable):
    name = TemplateColumn(
        verbose_name=_("Name"),
        template_name="oscar/dashboard/catalogue/option_row_name.html",
        order_by="name",
    )
    actions = TemplateColumn(
        verbose_name=_("Actions"),
        template_name="oscar/dashboard/catalogue/option_row_actions.html",
        orderable=False,
    )

    icon = "reorder"
    caption = ngettext_lazy("%s Option", "%s Options")

    class Meta(DashboardTable.Meta):
        model = Option
        fields = ("name", "type", "required")
        sequence = ("name", "type", "required", "actions")
        per_page = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE



# coding=utf-8
import factory

from oscar.core.loading import get_model

__all__ = [
    "ProductClassFactory",
    "ProductFactory",
    "CategoryFactory",
    "ProductCategoryFactory",
    "ProductAttributeFactory",
    "AttributeOptionGroupFactory",
    "OptionFactory",
    "AttributeOptionFactory",
    "ProductAttributeValueFactory",
    "ProductReviewFactory",
    "ProductImageFactory",
]


class ProductClassFactory(factory.django.DjangoModelFactory):
    name = "Books"
    requires_shipping = True
    track_stock = True

    class Meta:
        model = get_model("catalogue", "ProductClass")


class ProductFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_model("catalogue", "Product")

    structure = Meta.model.STANDALONE
    upc = factory.Sequence(lambda n: "978080213020%d" % n)
    title = "A confederacy of dunces"
    product_class = factory.SubFactory(ProductClassFactory)

    stockrecords = factory.RelatedFactory(
        "oscar.test.factories.StockRecordFactory", "product"
    )
    categories = factory.RelatedFactory(
        "oscar.test.factories.ProductCategoryFactory", "product"
    )


class CategoryFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Category %d" % n)

    # Very naive handling of treebeard node fields. Works though!
    depth = 1
    path = factory.Sequence(lambda n: "%04d" % n)

    class Meta:
        model = get_model("catalogue", "Category")


class ProductCategoryFactory(factory.django.DjangoModelFactory):
    category = factory.SubFactory(CategoryFactory)

    class Meta:
        model = get_model("catalogue", "ProductCategory")


class ProductAttributeFactory(factory.django.DjangoModelFactory):
    code = name = "weight"
    product_class = factory.SubFactory(ProductClassFactory)
    type = "float"

    class Meta:
        model = get_model("catalogue", "ProductAttribute")


class OptionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_model("catalogue", "Option")

    name = "example option"
    code = "example"
    type = Meta.model.TEXT
    required = False


class AttributeOptionFactory(factory.django.DjangoModelFactory):
    # Ideally we'd get_or_create a AttributeOptionGroup here, but I'm not
    # aware of how to not create a unique option group for each call of the
    # factory

    option = factory.Sequence(lambda n: "Option %d" % n)

    class Meta:
        model = get_model("catalogue", "AttributeOption")


class AttributeOptionGroupFactory(factory.django.DjangoModelFactory):
    name = "Grüppchen"

    class Meta:
        model = get_model("catalogue", "AttributeOptionGroup")


class ProductAttributeValueFactory(factory.django.DjangoModelFactory):
    attribute = factory.SubFactory(ProductAttributeFactory)
    product = factory.SubFactory(ProductFactory)

    class Meta:
        model = get_model("catalogue", "ProductAttributeValue")


class ProductReviewFactory(factory.django.DjangoModelFactory):
    score = 5
    product = factory.SubFactory(ProductFactory, stockrecords=[])

    class Meta:
        model = get_model("reviews", "ProductReview")


class ProductImageFactory(factory.django.DjangoModelFactory):
    product = factory.SubFactory(ProductFactory, stockrecords=[])
    original = factory.django.ImageField(
        width=100, height=200, filename="test_image.jpg"
    )

    class Meta:
        model = get_model("catalogue", "ProductImage")



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('customer', '0004_email_save'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='date_sent',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
    ]



# Generated by Django 3.2.19 on 2023-07-10 08:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('basket', '0011_json_basket_option'),
    ]

    operations = [
        migrations.AddField(
            model_name='line',
            name='tax_code',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='VAT rate code'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-11-20 17:07
from __future__ import unicode_literals

from django.db import migrations
import oscar.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0002_auto_20151210_1053'),
    ]

    operations = [
        migrations.AlterField(
            model_name='condition',
            name='proxy_class',
            field=oscar.models.fields.NullCharField(default=None, max_length=255, verbose_name='Custom class'),
        ),
    ]



from oscar.apps.address.abstract_models import AbstractCountry, AbstractUserAddress
from oscar.core.loading import is_model_registered

__all__ = []


if not is_model_registered("address", "UserAddress"):

    class UserAddress(AbstractUserAddress):
        pass

    __all__.append("UserAddress")


if not is_model_registered("address", "Country"):

    class Country(AbstractCountry):
        pass

    __all__.append("Country")



from django.db.models import OuterRef, Subquery, Sum
from django.db.models.functions import Coalesce
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ReportGenerator = get_class("dashboard.reports.reports", "ReportGenerator")
ReportCSVFormatter = get_class("dashboard.reports.reports", "ReportCSVFormatter")
ReportHTMLFormatter = get_class("dashboard.reports.reports", "ReportHTMLFormatter")
ConditionalOffer = get_model("offer", "ConditionalOffer")
OrderDiscount = get_model("order", "OrderDiscount")


class OfferReportCSVFormatter(ReportCSVFormatter):
    filename_template = "conditional-offer-performance.csv"

    def generate_csv(self, response, offer_discounts):
        writer = self.get_csv_writer(response)
        header_row = [_("Offer"), _("Total discount")]
        writer.writerow(header_row)

        for discount in offer_discounts:
            writer.writerow(
                [discount["display_offer_name"], discount["total_discount"]]
            )


class OfferReportHTMLFormatter(ReportHTMLFormatter):
    filename_template = "oscar/dashboard/reports/partials/offer_report.html"


class OfferReportGenerator(ReportGenerator):
    code = "conditional-offers"
    description = _("Offer performance")
    model_class = OrderDiscount

    formatters = {
        "CSV_formatter": OfferReportCSVFormatter,
        "HTML_formatter": OfferReportHTMLFormatter,
    }

    def get_queryset(self):
        offers = ConditionalOffer.objects.filter(pk=OuterRef("offer_id"))
        return (
            super()
            .get_queryset()
            .order_by()
            .values("offer_id")
            .annotate(
                total_discount=Sum("amount"),
                # Used to add a link to the offer in the report template, if the offer still exists.
                offer=Subquery(offers.values("pk")[:1]),
                # Find the name of the attached offer if it exists, otherwise the offer_name on a matching OrderDiscount
                # This is used to display the most appropriate name in the report template.
                display_offer_name=Coalesce(
                    Subquery(offers.values("name")[:1]),
                    Subquery(
                        OrderDiscount.objects.filter(pk=OuterRef("pk")).values(
                            "offer_name"
                        )[:1]
                    ),
                ),
            )
            .values("offer_id", "offer", "total_discount", "display_offer_name")
            .order_by("-total_discount")
        )



import logging

from django.core.management.base import BaseCommand, CommandError

from oscar.core.loading import get_class

CatalogueImporter = get_class("partner.importers", "CatalogueImporter")
ImportingError = get_class("partner.exceptions", "ImportingError")

logger = logging.getLogger("oscar.catalogue.import")


class Command(BaseCommand):
    help = "For creating product catalogues based on a CSV file"

    def add_arguments(self, parser):
        parser.add_argument(
            "filename", nargs="+", help="/path/to/file1.csv /path/to/file2.csv ..."
        )
        parser.add_argument(
            "--flush",
            action="store_true",
            dest="flush",
            default=False,
            help="Flush tables before importing",
        )
        parser.add_argument(
            "--delimiter",
            dest="delimiter",
            default=",",
            help="Delimiter used within CSV file(s)",
        )

    def handle(self, *args, **options):
        logger.info("Starting catalogue import")
        importer = CatalogueImporter(
            logger, delimiter=options.get("delimiter"), flush=options.get("flush")
        )
        for file_path in options["filename"]:
            logger.info(" - Importing records from '%s'", file_path)
            try:
                importer.handle(file_path)
            except ImportingError as e:
                raise CommandError(str(e))



from django.forms import TextInput, fields

from oscar.core import validators


class ExtendedURLField(fields.URLField):
    """
    Custom field similar to URLField type field, however also accepting and
    validating local relative URLs, ie. '/product/'
    """

    default_validators = []
    # Django 1.6 renders URLInput as <input type=url>, which causes some
    # browsers to require the input to be a valid absolute URL. As relative
    # URLS are allowed for ExtendedURLField, we must set it to TextInput
    widget = TextInput

    def __init__(self, *args, **kwargs):
        super(fields.URLField, self).__init__(*args, **kwargs)
        self.validators.append(validators.ExtendedURLValidator())

    def to_python(self, value):
        # We need to avoid having 'http' inserted at the start of
        # every value so that local URLs are valid.
        if value and value.startswith("/"):
            return value
        return super().to_python(value)



import csv
import os
from decimal import Decimal as D

from django.db.transaction import atomic
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model

ImportingError = get_class("partner.exceptions", "ImportingError")

Category = get_model("catalogue", "Category")
Partner = get_model("partner", "Partner")
Product = get_model("catalogue", "Product")
ProductCategory = get_model("catalogue", "ProductCategory")
ProductClass = get_model("catalogue", "ProductClass")
StockRecord = get_model("partner", "StockRecord")

create_from_breadcrumbs = get_class("catalogue.categories", "create_from_breadcrumbs")


class CatalogueImporter(object):
    """
    CSV product importer used to built sandbox. Might not work very well
    for anything else.
    """

    _flush = False

    def __init__(self, logger, delimiter=",", flush=False):
        self.logger = logger
        self._delimiter = delimiter
        self._flush = flush

    def handle(self, file_path=None):
        """Handles the actual import process"""
        if not file_path:
            raise ImportingError(_("No file path supplied"))
        Validator().validate(file_path)
        if self._flush is True:
            self.logger.info(" - Flushing product data before import")
            self._flush_product_data()
        self._import(file_path)

    def _flush_product_data(self):
        """Flush out product and stock models"""
        Product.objects.all().delete()
        ProductClass.objects.all().delete()
        Partner.objects.all().delete()
        StockRecord.objects.all().delete()

    @atomic
    def _import(self, file_path):
        """Imports given file"""
        stats = {"new_items": 0, "updated_items": 0}
        row_number = 0
        with open(file_path, "rt", encoding="utf-8") as f:
            reader = csv.reader(f, escapechar="\\")
            for row in reader:
                row_number += 1
                self._import_row(row_number, row, stats)
        msg = "New items: %d, updated items: %d" % (
            stats["new_items"],
            stats["updated_items"],
        )
        self.logger.info(msg)

    def _import_row(self, row_number, row, stats):
        if len(row) != 5 and len(row) != 9:
            self.logger.error(
                "Row number %d has an invalid number of fields"
                " (%d), skipping..." % (row_number, len(row))
            )
            return
        item = self._create_item(*row[:5], stats=stats)
        if len(row) == 9:
            # With stock data
            self._create_stockrecord(item, *row[5:9])

    def _create_item(self, product_class, category_str, upc, title, description, stats):
        # Ignore any entries that are NULL
        if description == "NULL":
            description = ""

        # Create item class and item
        product_class, __ = ProductClass.objects.get_or_create(name=product_class)
        try:
            item = Product.objects.get(upc=upc)
            stats["updated_items"] += 1
        except Product.DoesNotExist:
            item = Product()
            stats["new_items"] += 1
        item.upc = upc
        item.title = title
        item.description = description
        item.product_class = product_class
        item.save()

        # Category
        cat = create_from_breadcrumbs(category_str)
        ProductCategory.objects.update_or_create(product=item, category=cat)

        return item

    def _create_stockrecord(self, item, partner_name, partner_sku, price, num_in_stock):
        # Create partner and stock record
        partner, _ = Partner.objects.get_or_create(name=partner_name)
        try:
            stock = StockRecord.objects.get(partner_sku=partner_sku)
        except StockRecord.DoesNotExist:
            stock = StockRecord()

        stock.product = item
        stock.partner = partner
        stock.partner_sku = partner_sku
        stock.price = D(price)
        stock.num_in_stock = num_in_stock
        stock.save()


class Validator(object):
    def validate(self, file_path):
        self._exists(file_path)
        self._is_file(file_path)
        self._is_readable(file_path)

    def _exists(self, file_path):
        """Check whether a file exists"""
        if not os.path.exists(file_path):
            raise ImportingError(_("%s does not exist") % (file_path))

    def _is_file(self, file_path):
        """Check whether file is actually a file type"""
        if not os.path.isfile(file_path):
            raise ImportingError(_("%s is not a file") % (file_path))

    def _is_readable(self, file_path):
        """Check file is readable"""
        try:
            f = open(file_path, "r", encoding="utf-8")
            f.close()
        except IOError:
            raise ImportingError(_("%s is not readable") % (file_path))



from django import template

from oscar.core.loading import get_model

register = template.Library()
Category = get_model("catalogue", "category")


class PassThrough(object):
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, objtype):
        if obj is None:
            return self

        return getattr(obj.category, self.name)


class CategoryFieldPassThroughMetaClass(type):
    """
    Add accessors for category fields to whichever class is of this type.
    """

    def __new__(cls, name, bases, attrs):
        field_accessors = {}
        for field in Category._meta.get_fields():
            name = field.name
            field_accessors[name] = PassThrough(name)

        # attrs win of silly field accessors
        field_accessors.update(attrs)
        return type.__new__(cls, name, bases, field_accessors)


class CheapCategoryInfo(dict, metaclass=CategoryFieldPassThroughMetaClass):
    """
    Wrapper class for Category.

    Besides allowing inclusion of extra info, useful while rendering a template,
    this class hides any expensive properties people should not use by accident
    in templates.

    This replaces both the node as the info object returned by the ``category_tree``
    templatetag, so it mimics a tuple of 2 items (which are the same) for
    backwards compatibility.
    """

    def __init__(self, category, **info):
        super().__init__(info)
        self.category = category

    @property
    def pk(self):
        return self.category.pk

    def get_absolute_url(self):
        return self["url"]

    def __len__(self):
        "Mimic a tuple of 2 items"
        return 2

    def __iter__(self):
        "be an iterable of 2 times the same item"
        yield self
        yield self


@register.simple_tag(name="category_tree")
def get_annotated_list(depth=None, parent=None):
    """
    Gets an annotated list from a tree branch.

    Borrows heavily from treebeard's get_annotated_list
    """
    # 'depth' is the backwards-compatible name for the template tag,
    # 'max_depth' is the better variable name.
    max_depth = depth

    annotated_categories = []
    tree_slug = ""

    start_depth, prev_depth = (None, None)
    if parent:
        categories = parent.get_descendants()
        tree_slug = parent.get_full_slug()
        if max_depth is not None:
            max_depth += parent.get_depth()
    else:
        categories = Category.get_tree()

    if max_depth is not None:
        categories = categories.filter(depth__lte=max_depth)

    categories = categories.for_menu()

    info = CheapCategoryInfo(parent, url="")

    node_depth = 0
    for node in categories:
        node_depth = node.get_depth()
        if start_depth is None:
            start_depth = node_depth

        # Update previous node's info
        if prev_depth is None or node_depth > prev_depth:
            info["has_children"] = True
            if info.category is not None:
                tree_slug = info.category.get_full_slug(tree_slug)

        if prev_depth is not None and node_depth < prev_depth:
            depth_difference = prev_depth - node_depth
            info["num_to_close"] = list(range(0, depth_difference))
            tree_slugs = tree_slug.rsplit(node._slug_separator, depth_difference)
            if tree_slugs:
                tree_slug = tree_slugs[0]
            else:
                tree_slug = node.slug

        info = CheapCategoryInfo(
            node,
            url=node._get_absolute_url(tree_slug),
            num_to_close=[],
            level=node_depth - start_depth,
        )
        annotated_categories.append(info)

        prev_depth = node_depth

    if prev_depth is not None:
        # close last leaf
        info["num_to_close"] = list(range(0, prev_depth - start_depth))
        info["has_children"] = node_depth > prev_depth

    return annotated_categories



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import oscar.core.validators
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ProductReview',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('score', models.SmallIntegerField(verbose_name='Score', choices=[(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)])),
                ('title', models.CharField(max_length=255, verbose_name='Title', validators=[oscar.core.validators.non_whitespace])),
                ('body', models.TextField(verbose_name='Body')),
                ('name', models.CharField(max_length=255, verbose_name='Name', blank=True)),
                ('email', models.EmailField(max_length=75, verbose_name='Email', blank=True)),
                ('homepage', models.URLField(verbose_name='URL', blank=True)),
                ('status', models.SmallIntegerField(default=1, verbose_name='Status', choices=[(0, 'Requires moderation'), (1, 'Approved'), (2, 'Rejected')])),
                ('total_votes', models.IntegerField(default=0, verbose_name='Total Votes')),
                ('delta_votes', models.IntegerField(default=0, db_index=True, verbose_name='Delta Votes')),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, related_name='reviews', to='catalogue.Product', null=True)),
                ('user', models.ForeignKey(null=True, related_name='reviews', to=settings.AUTH_USER_MODEL, blank=True, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-delta_votes', 'id'],
                'verbose_name_plural': 'Product reviews',
                'verbose_name': 'Product review',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('delta', models.SmallIntegerField(verbose_name='Delta', choices=[(1, 'Up'), (-1, 'Down')])),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('review', models.ForeignKey(related_name='votes', to='reviews.ProductReview', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(related_name='review_votes', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-date_created'],
                'verbose_name_plural': 'Votes',
                'verbose_name': 'Vote',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='vote',
            unique_together=set([('user', 'review')]),
        ),
        migrations.AlterUniqueTogether(
            name='productreview',
            unique_together=set([('product', 'user')]),
        ),
    ]



from django import forms
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.core.compat import get_user_model
from oscar.core.loading import get_model

User = get_user_model()
ProductAlert = get_model("customer", "ProductAlert")


class UserSearchForm(forms.Form):
    email = forms.CharField(required=False, label=_("Email"))
    name = forms.CharField(required=False, label=pgettext_lazy("User's name", "Name"))


class ProductAlertUpdateForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        alert = kwargs["instance"]
        if alert.user:
            # Remove 'unconfirmed' from list of available choices when editing
            # an alert for a real user
            choices = self.fields["status"].choices
            del choices[0]
            self.fields["status"].choices = choices

    class Meta:
        model = ProductAlert
        fields = [
            "status",
        ]


class ProductAlertSearchForm(forms.Form):
    STATUS_CHOICES = (("", "------------"),) + ProductAlert.STATUS_CHOICES

    status = forms.ChoiceField(
        required=False, choices=STATUS_CHOICES, label=_("Status")
    )
    name = forms.CharField(required=False, label=_("Name"))
    email = forms.EmailField(required=False, label=_("Email"))



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class ShippingConfig(OscarConfig):
    label = "shipping"
    name = "oscar.apps.shipping"
    verbose_name = _("Shipping")



# Generated by Django 2.0.3 on 2018-04-02 13:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0004_auto_20180228_0940'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voucherset',
            name='offer',
            field=models.OneToOneField(blank=True, limit_choices_to={'offer_type': 'Voucher'}, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='voucher_set', to='offer.ConditionalOffer', verbose_name='Offer'),
        ),
    ]



from django import forms
from django.db import transaction
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from oscar.apps.voucher.utils import get_unused_code
from oscar.core.loading import get_model
from oscar.forms import widgets

Voucher = get_model("voucher", "Voucher")
VoucherSet = get_model("voucher", "VoucherSet")
ConditionalOffer = get_model("offer", "ConditionalOffer")


class VoucherForm(forms.ModelForm):
    """
    A specialised form for creating a voucher model, and capturing the offers
    that apply to it.
    """

    offers = forms.ModelMultipleChoiceField(
        label=_("Which offers apply for this voucher?"),
        queryset=ConditionalOffer.objects.filter(offer_type=ConditionalOffer.VOUCHER),
    )

    class Meta:
        model = Voucher
        fields = [
            "name",
            "code",
            "start_datetime",
            "end_datetime",
            "usage",
        ]
        widgets = {
            "start_datetime": widgets.DateTimePickerInput(),
            "end_datetime": widgets.DateTimePickerInput(),
        }

    def clean_code(self):
        return self.cleaned_data["code"].strip().upper()


class VoucherSearchForm(forms.Form):
    name = forms.CharField(required=False, label=_("Name"))
    code = forms.CharField(required=False, label=_("Code"))
    offer_name = forms.CharField(required=False, label=_("Offer name"))
    is_active = forms.NullBooleanField(
        required=False, label=_("Is active?"), widget=widgets.NullBooleanSelect
    )
    in_set = forms.NullBooleanField(
        required=False, label=_("In voucher set?"), widget=widgets.NullBooleanSelect
    )
    has_offers = forms.NullBooleanField(
        required=False, label=_("Has offers?"), widget=widgets.NullBooleanSelect
    )

    basic_fields = [
        "name",
        "code",
        "is_active",
        "in_set",
    ]

    def clean_code(self):
        return self.cleaned_data["code"].upper()


class VoucherSetForm(forms.ModelForm):
    usage = forms.ChoiceField(
        choices=(("", "---------"),) + Voucher.USAGE_CHOICES, label=_("Usage")
    )

    offers = forms.ModelMultipleChoiceField(
        label=_("Which offers apply for this voucher set?"),
        queryset=ConditionalOffer.objects.filter(offer_type=ConditionalOffer.VOUCHER),
    )

    class Meta:
        model = VoucherSet
        fields = [
            "name",
            "code_length",
            "description",
            "start_datetime",
            "end_datetime",
            "count",
        ]
        widgets = {
            "start_datetime": widgets.DateTimePickerInput(),
            "end_datetime": widgets.DateTimePickerInput(),
        }

    def clean_count(self):
        data = self.cleaned_data["count"]
        if (self.instance.pk is not None) and (data < self.instance.count):
            detail_url = reverse(
                "dashboard:voucher-set-detail", kwargs={"pk": self.instance.pk}
            )
            raise forms.ValidationError(
                mark_safe(
                    _(
                        "This cannot be used to delete vouchers (currently %s) in this set. "
                        'You can do that on the <a href="%s">detail</a> page.'
                    )
                    % (self.instance.count, detail_url)
                )
            )
        return data

    @transaction.atomic
    def save(self, commit=True):
        instance = super().save(commit)
        if commit:
            usage = self.cleaned_data["usage"]
            offers = self.cleaned_data["offers"]
            if instance is not None:
                # Update vouchers in this set
                for i, voucher in enumerate(instance.vouchers.order_by("date_created")):
                    voucher.name = "%s - %d" % (instance.name, i + 1)
                    voucher.usage = usage
                    voucher.start_datetime = instance.start_datetime
                    voucher.end_datetime = instance.end_datetime
                    voucher.save()
                    voucher.offers.set(offers)
            # Add vouchers to this set
            vouchers_added = False
            for i in range(instance.vouchers.count(), instance.count):
                voucher = Voucher.objects.create(
                    name="%s - %d" % (instance.name, i + 1),
                    code=get_unused_code(length=instance.code_length),
                    voucher_set=instance,
                    usage=usage,
                    start_datetime=instance.start_datetime,
                    end_datetime=instance.end_datetime,
                )
                voucher.offers.add(*offers)
                vouchers_added = True
            if vouchers_added:
                instance.update_count()
        return instance


class VoucherSetSearchForm(forms.Form):
    code = forms.CharField(required=False, label=_("Code"))

    def clean_code(self):
        return self.cleaned_data["code"].upper()



import logging
from functools import lru_cache

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.urls import resolve, reverse

from oscar.core.application import OscarDashboardConfig
from oscar.views.decorators import check_permissions

logger = logging.getLogger("oscar.dashboard")


class Node(object):
    """
    A node in the dashboard navigation menu
    """

    def __init__(
        self,
        label,
        url_name=None,
        url_args=None,
        url_kwargs=None,
        access_fn=None,
        icon=None,
    ):
        self.label = label
        self.icon = icon
        self.url_name = url_name
        self.url_args = url_args
        self.url_kwargs = url_kwargs
        self.access_fn = access_fn
        self.children = []

    @property
    def is_heading(self):
        return self.url_name is None

    @property
    def url(self):
        return reverse(self.url_name, args=self.url_args, kwargs=self.url_kwargs)

    def add_child(self, node):
        self.children.append(node)

    def is_visible(self, user):
        return self.access_fn is None or self.access_fn(
            user, self.url_name, self.url_args, self.url_kwargs
        )

    def filter(self, user):
        if not self.is_visible(user):
            return None
        node = Node(
            label=self.label,
            url_name=self.url_name,
            url_args=self.url_args,
            url_kwargs=self.url_kwargs,
            access_fn=self.access_fn,
            icon=self.icon,
        )
        for child in self.children:
            if child.is_visible(user):
                node.add_child(child)
        return node

    def has_children(self):
        return len(self.children) > 0


@lru_cache(maxsize=1)
def _dashboard_url_names_to_config():
    dashboard_configs = (
        config
        for config in apps.get_app_configs()
        if isinstance(config, OscarDashboardConfig)
    )
    urls_to_config = {}
    for config in dashboard_configs:
        for url in config.urls[0]:
            # includes() don't have a name attribute
            # We skipped them because they come from other AppConfigs
            name = getattr(url, "name", None)
            if not name:
                continue

            if name in urls_to_config:
                if urls_to_config[name] != config:
                    raise ImproperlyConfigured(
                        "'{}' exists in both {} and {}!".format(
                            name, config, urls_to_config[name]
                        )
                    )

            urls_to_config[name] = config
    return urls_to_config


def default_access_fn(user, url_name, url_args=None, url_kwargs=None):
    """
    Given a user and a url_name, this function assesses whether the
    user has the right to access the URL.
    Once the permissions for the view are known, the access logic used
    by the dashboard decorator is evaluated
    """
    if url_name is None:  # it's a heading
        return True

    url = reverse(url_name, args=url_args, kwargs=url_kwargs)
    url_match = resolve(url)
    url_name = url_match.url_name
    app_config_instance = _dashboard_url_names_to_config()[url_name]

    permissions = app_config_instance.get_permissions(url_name)

    return check_permissions(user, permissions)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-26 11:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('address', '0003_auto_20150927_1551'),
    ]

    operations = [
        migrations.AlterField(
            model_name='useraddress',
            name='num_orders_as_billing_address',
            field=models.PositiveIntegerField(default=0, verbose_name='Number of Orders as Billing Address'),
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='num_orders_as_shipping_address',
            field=models.PositiveIntegerField(default=0, verbose_name='Number of Orders as Shipping Address'),
        ),
    ]



from decimal import Decimal as D

from django.utils.translation import gettext_lazy as _

from oscar.core import prices


class BaseSurcharge:
    """
    Surcharge interface class

    This is the superclass to the classes in surcharges.py. This allows using all
    surcharges interchangeably (aka polymorphism).

    The interface is all properties.
    """

    def calculate(self, basket, **kwargs):
        raise NotImplementedError


class PercentageCharge(BaseSurcharge):
    name = _("Percentage surcharge")
    code = "percentage-surcharge"

    def __init__(self, percentage):
        self.percentage = percentage

    def calculate(self, basket, **kwargs):
        if not basket.is_empty:
            shipping_charge = kwargs.get("shipping_charge")

            if shipping_charge is not None:
                total_excl_tax = basket.total_excl_tax + shipping_charge.excl_tax
                total_incl_tax = basket.total_incl_tax + shipping_charge.incl_tax
            else:
                total_excl_tax = basket.total_excl_tax
                total_incl_tax = basket.total_incl_tax

            return prices.Price(
                currency=basket.currency,
                excl_tax=total_excl_tax * self.percentage / 100,
                incl_tax=total_incl_tax * self.percentage / 100,
            )
        else:
            return prices.Price(
                currency=basket.currency, excl_tax=D("0.0"), incl_tax=D("0.0")
            )


class FlatCharge(BaseSurcharge):
    name = _("Flat surcharge")
    code = "flat-surcharge"

    def __init__(self, excl_tax=None, incl_tax=None):
        self.excl_tax = excl_tax
        self.incl_tax = incl_tax

    def calculate(self, basket, **kwargs):
        return prices.Price(
            currency=basket.currency, excl_tax=self.excl_tax, incl_tax=self.incl_tax
        )



# Generated by Django 2.2.9 on 2019-12-20 09:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0017_auto_20190816_0938'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='ancestors_are_public',
            field=models.BooleanField(db_index=True, default=True, help_text='The ancestors of this category are public', verbose_name='Ancestor categories are public'),
        ),
        migrations.AddField(
            model_name='category',
            name='is_public',
            field=models.BooleanField(db_index=True, default=True, help_text='Show this category in search results and catalogue listings.', verbose_name='Is public'),
        ),
        migrations.AlterField(
            model_name='product',
            name='is_public',
            field=models.BooleanField(db_index=True, default=True, help_text='Show this product in search results and catalogue listings.', verbose_name='Is public'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
import django.db.models.deletion
import oscar.models.fields
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('partner', '0001_initial'),
        ('customer', '0001_initial'),
        ('address', '0001_initial'),
        ('basket', '0002_auto_20140827_1705'),
        ('catalogue', '0001_initial'),
        ('sites', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BillingAddress',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(verbose_name='Title', max_length=64, blank=True, choices=[('Mr', 'Mr'), ('Miss', 'Miss'), ('Mrs', 'Mrs'), ('Ms', 'Ms'), ('Dr', 'Dr')])),
                ('first_name', models.CharField(max_length=255, verbose_name='First name', blank=True)),
                ('last_name', models.CharField(max_length=255, verbose_name='Last name', blank=True)),
                ('line1', models.CharField(max_length=255, verbose_name='First line of address')),
                ('line2', models.CharField(max_length=255, verbose_name='Second line of address', blank=True)),
                ('line3', models.CharField(max_length=255, verbose_name='Third line of address', blank=True)),
                ('line4', models.CharField(max_length=255, verbose_name='City', blank=True)),
                ('state', models.CharField(max_length=255, verbose_name='State/County', blank=True)),
                ('postcode', oscar.models.fields.UppercaseCharField(max_length=64, verbose_name='Post/Zip-code', blank=True)),
                ('search_text', models.TextField(editable=False, verbose_name='Search text - used only for searching addresses')),
                ('country', models.ForeignKey(verbose_name='Country', to='address.Country', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Billing addresses',
                'verbose_name': 'Billing address',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='CommunicationEvent',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date')),
                ('event_type', models.ForeignKey(verbose_name='Event Type', to='customer.CommunicationEventType', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-date_created'],
                'verbose_name_plural': 'Communication Events',
                'verbose_name': 'Communication Event',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Line',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('partner_name', models.CharField(max_length=128, verbose_name='Partner name', blank=True)),
                ('partner_sku', models.CharField(max_length=128, verbose_name='Partner SKU')),
                ('partner_line_reference', models.CharField(verbose_name='Partner reference', max_length=128, help_text='This is the item number that the partner uses within their system', blank=True)),
                ('partner_line_notes', models.TextField(verbose_name='Partner Notes', blank=True)),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('upc', models.CharField(verbose_name='UPC', max_length=128, blank=True, null=True)),
                ('quantity', models.PositiveIntegerField(default=1, verbose_name='Quantity')),
                ('line_price_incl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price (inc. tax)')),
                ('line_price_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price (excl. tax)')),
                ('line_price_before_discounts_incl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price before discounts (inc. tax)')),
                ('line_price_before_discounts_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price before discounts (excl. tax)')),
                ('unit_cost_price', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Unit Cost Price', null=True)),
                ('unit_price_incl_tax', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Unit Price (inc. tax)', null=True)),
                ('unit_price_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Unit Price (excl. tax)', null=True)),
                ('unit_retail_price', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Unit Retail Price', null=True)),
                ('status', models.CharField(max_length=255, verbose_name='Status', blank=True)),
                ('est_dispatch_date', models.DateField(blank=True, verbose_name='Estimated Dispatch Date', null=True)),
            ],
            options={
                'verbose_name_plural': 'Order Lines',
                'verbose_name': 'Order Line',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='LineAttribute',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(max_length=128, verbose_name='Type')),
                ('value', models.CharField(max_length=255, verbose_name='Value')),
                ('line', models.ForeignKey(verbose_name='Line', related_name='attributes', to='order.Line', on_delete=models.CASCADE)),
                ('option', models.ForeignKey(verbose_name='Option', on_delete=django.db.models.deletion.SET_NULL, related_name='line_attributes', to='catalogue.Option', null=True)),
            ],
            options={
                'verbose_name_plural': 'Line Attributes',
                'verbose_name': 'Line Attribute',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='LinePrice',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField(default=1, verbose_name='Quantity')),
                ('price_incl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price (inc. tax)')),
                ('price_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Price (excl. tax)')),
                ('shipping_incl_tax', models.DecimalField(default=0, max_digits=12, decimal_places=2, verbose_name='Shiping (inc. tax)')),
                ('shipping_excl_tax', models.DecimalField(default=0, max_digits=12, decimal_places=2, verbose_name='Shipping (excl. tax)')),
                ('line', models.ForeignKey(verbose_name='Line', related_name='prices', to='order.Line', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('id',),
                'verbose_name_plural': 'Line Prices',
                'verbose_name': 'Line Price',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('number', models.CharField(max_length=128, unique=True, db_index=True, verbose_name='Order number')),
                ('currency', models.CharField(default='GBP', max_length=12, verbose_name='Currency')),
                ('total_incl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Order total (inc. tax)')),
                ('total_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Order total (excl. tax)')),
                ('shipping_incl_tax', models.DecimalField(default=0, max_digits=12, decimal_places=2, verbose_name='Shipping charge (inc. tax)')),
                ('shipping_excl_tax', models.DecimalField(default=0, max_digits=12, decimal_places=2, verbose_name='Shipping charge (excl. tax)')),
                ('shipping_method', models.CharField(max_length=128, verbose_name='Shipping method', blank=True)),
                ('shipping_code', models.CharField(default='', max_length=128, blank=True)),
                ('status', models.CharField(max_length=100, verbose_name='Status', blank=True)),
                ('guest_email', models.EmailField(max_length=75, verbose_name='Guest email address', blank=True)),
                ('date_placed', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('basket', models.ForeignKey(null=True, verbose_name='Basket', on_delete=django.db.models.deletion.SET_NULL, to='basket.Basket', blank=True)),
                ('billing_address', models.ForeignKey(null=True, verbose_name='Billing Address', on_delete=django.db.models.deletion.SET_NULL, to='order.BillingAddress', blank=True)),
            ],
            options={
                'ordering': ['-date_placed'],
                'verbose_name_plural': 'Orders',
                'verbose_name': 'Order',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='OrderDiscount',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(default='Basket', max_length=64, verbose_name='Discount category', choices=[('Basket', 'Basket'), ('Shipping', 'Shipping'), ('Deferred', 'Deferred')])),
                ('offer_id', models.PositiveIntegerField(blank=True, verbose_name='Offer ID', null=True)),
                ('offer_name', models.CharField(max_length=128, db_index=True, verbose_name='Offer name', blank=True)),
                ('voucher_id', models.PositiveIntegerField(blank=True, verbose_name='Voucher ID', null=True)),
                ('voucher_code', models.CharField(max_length=128, db_index=True, verbose_name='Code', blank=True)),
                ('frequency', models.PositiveIntegerField(verbose_name='Frequency', null=True)),
                ('amount', models.DecimalField(default=0, max_digits=12, decimal_places=2, verbose_name='Amount')),
                ('message', models.TextField(blank=True)),
                ('order', models.ForeignKey(verbose_name='Order', related_name='discounts', to='order.Order', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Order Discounts',
                'verbose_name': 'Order Discount',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='OrderNote',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('note_type', models.CharField(max_length=128, verbose_name='Note Type', blank=True)),
                ('message', models.TextField(verbose_name='Message')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('order', models.ForeignKey(verbose_name='Order', related_name='notes', to='order.Order', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(verbose_name='User', to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Order Notes',
                'verbose_name': 'Order Note',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PaymentEvent',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Amount')),
                ('reference', models.CharField(max_length=128, verbose_name='Reference', blank=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
            ],
            options={
                'ordering': ['-date_created'],
                'verbose_name_plural': 'Payment Events',
                'verbose_name': 'Payment Event',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PaymentEventQuantity',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField(verbose_name='Quantity')),
                ('event', models.ForeignKey(verbose_name='Event', related_name='line_quantities', to='order.PaymentEvent', on_delete=models.CASCADE)),
                ('line', models.ForeignKey(verbose_name='Line', related_name='payment_event_quantities', to='order.Line', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Payment Event Quantities',
                'verbose_name': 'Payment Event Quantity',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PaymentEventType',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=128, verbose_name='Name')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Code', max_length=128, editable=False, blank=True)),
            ],
            options={
                'ordering': ('name',),
                'verbose_name_plural': 'Payment Event Types',
                'verbose_name': 'Payment Event Type',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ShippingAddress',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(verbose_name='Title', max_length=64, blank=True, choices=[('Mr', 'Mr'), ('Miss', 'Miss'), ('Mrs', 'Mrs'), ('Ms', 'Ms'), ('Dr', 'Dr')])),
                ('first_name', models.CharField(max_length=255, verbose_name='First name', blank=True)),
                ('last_name', models.CharField(max_length=255, verbose_name='Last name', blank=True)),
                ('line1', models.CharField(max_length=255, verbose_name='First line of address')),
                ('line2', models.CharField(max_length=255, verbose_name='Second line of address', blank=True)),
                ('line3', models.CharField(max_length=255, verbose_name='Third line of address', blank=True)),
                ('line4', models.CharField(max_length=255, verbose_name='City', blank=True)),
                ('state', models.CharField(max_length=255, verbose_name='State/County', blank=True)),
                ('postcode', oscar.models.fields.UppercaseCharField(max_length=64, verbose_name='Post/Zip-code', blank=True)),
                ('search_text', models.TextField(editable=False, verbose_name='Search text - used only for searching addresses')),
                ('phone_number', oscar.models.fields.PhoneNumberField(verbose_name='Phone number', help_text='In case we need to call you about your order', blank=True)),
                ('notes', models.TextField(verbose_name='Instructions', help_text='Tell us anything we should know when delivering your order.', blank=True)),
                ('country', models.ForeignKey(verbose_name='Country', to='address.Country', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Shipping addresses',
                'verbose_name': 'Shipping address',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ShippingEvent',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('notes', models.TextField(verbose_name='Event notes', help_text='This could be the dispatch reference, or a tracking number', blank=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
            ],
            options={
                'ordering': ['-date_created'],
                'verbose_name_plural': 'Shipping Events',
                'verbose_name': 'Shipping Event',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ShippingEventQuantity',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField(verbose_name='Quantity')),
                ('event', models.ForeignKey(verbose_name='Event', related_name='line_quantities', to='order.ShippingEvent', on_delete=models.CASCADE)),
                ('line', models.ForeignKey(verbose_name='Line', related_name='shipping_event_quantities', to='order.Line', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Shipping Event Quantities',
                'verbose_name': 'Shipping Event Quantity',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ShippingEventType',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=255, verbose_name='Name')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Code', max_length=128, editable=False, blank=True)),
            ],
            options={
                'ordering': ('name',),
                'verbose_name_plural': 'Shipping Event Types',
                'verbose_name': 'Shipping Event Type',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='shippingeventquantity',
            unique_together=set([('event', 'line')]),
        ),
        migrations.AddField(
            model_name='shippingevent',
            name='event_type',
            field=models.ForeignKey(verbose_name='Event Type', to='order.ShippingEventType', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='shippingevent',
            name='lines',
            field=models.ManyToManyField(related_name='shipping_events', verbose_name='Lines', to='order.Line', through='order.ShippingEventQuantity'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='shippingevent',
            name='order',
            field=models.ForeignKey(verbose_name='Order', related_name='shipping_events', to='order.Order', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='paymenteventquantity',
            unique_together=set([('event', 'line')]),
        ),
        migrations.AddField(
            model_name='paymentevent',
            name='event_type',
            field=models.ForeignKey(verbose_name='Event Type', to='order.PaymentEventType', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='paymentevent',
            name='lines',
            field=models.ManyToManyField(through='order.PaymentEventQuantity', verbose_name='Lines', to='order.Line'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='paymentevent',
            name='order',
            field=models.ForeignKey(verbose_name='Order', related_name='payment_events', to='order.Order', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='paymentevent',
            name='shipping_event',
            field=models.ForeignKey(related_name='payment_events', to='order.ShippingEvent', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='order',
            name='shipping_address',
            field=models.ForeignKey(null=True, verbose_name='Shipping Address', on_delete=django.db.models.deletion.SET_NULL, to='order.ShippingAddress', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='order',
            name='site',
            field=models.ForeignKey(verbose_name='Site', on_delete=django.db.models.deletion.SET_NULL, to='sites.Site', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='order',
            name='user',
            field=models.ForeignKey(null=True, verbose_name='User', on_delete=django.db.models.deletion.SET_NULL, related_name='orders', to=settings.AUTH_USER_MODEL, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='lineprice',
            name='order',
            field=models.ForeignKey(verbose_name='Order', related_name='line_prices', to='order.Order', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='order',
            field=models.ForeignKey(verbose_name='Order', related_name='lines', to='order.Order', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='partner',
            field=models.ForeignKey(null=True, verbose_name='Partner', on_delete=django.db.models.deletion.SET_NULL, related_name='order_lines', to='partner.Partner', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='product',
            field=models.ForeignKey(null=True, verbose_name='Product', on_delete=django.db.models.deletion.SET_NULL, to='catalogue.Product', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='stockrecord',
            field=models.ForeignKey(null=True, verbose_name='Stock record', on_delete=django.db.models.deletion.SET_NULL, to='partner.StockRecord', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='communicationevent',
            name='order',
            field=models.ForeignKey(verbose_name='Order', related_name='communication_events', to='order.Order', on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-07-04 07:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0006_auto_20170504_0616'),
    ]

    operations = [
        migrations.AddField(
            model_name='conditionaloffer',
            name='exclusive',
            field=models.BooleanField(default=True, help_text='Exclusive offers cannot be combined on the same items', verbose_name='Exclusive offer'),
        ),
    ]



import factory

from oscar.core.loading import get_class, get_model

Selector = get_class("partner.strategy", "Selector")


__all__ = ["BasketFactory", "BasketLineAttributeFactory"]


class BasketFactory(factory.django.DjangoModelFactory):
    # pylint: disable=unused-argument, W0201
    @factory.post_generation
    def set_strategy(self, create, extracted, **kwargs):
        # Load default strategy (without a user/request)
        self.strategy = Selector().strategy()

    class Meta:
        model = get_model("basket", "Basket")


class BasketLineAttributeFactory(factory.django.DjangoModelFactory):
    option = factory.SubFactory("oscar.test.factories.OptionFactory")

    class Meta:
        model = get_model("basket", "LineAttribute")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-13 08:44
from __future__ import unicode_literals

from django.db import migrations
import oscar.models.fields.slugfield


class Migration(migrations.Migration):

    dependencies = [
        ('basket', '0006_auto_20160111_1108'),
    ]

    operations = [
        migrations.AlterField(
            model_name='line',
            name='line_reference',
            field=oscar.models.fields.slugfield.SlugField(max_length=128, verbose_name='Line Reference'),
        ),
    ]



from django import shortcuts
from django.contrib import messages
from django.template.loader import render_to_string
from django.urls import reverse
from django.views import generic

from oscar.core.loading import get_classes, get_model

WeightBandForm, WeightBasedForm = get_classes(
    "dashboard.shipping.forms", ["WeightBandForm", "WeightBasedForm"]
)
WeightBased = get_model("shipping", "WeightBased")
WeightBand = get_model("shipping", "WeightBand")


class WeightBasedListView(generic.ListView):
    model = WeightBased
    template_name = "oscar/dashboard/shipping/weight_based_list.html"
    context_object_name = "methods"


class WeightBasedCreateView(generic.CreateView):
    model = WeightBased
    form_class = WeightBasedForm
    template_name = "oscar/dashboard/shipping/weight_based_form.html"

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/method_created.html",
            {"method": self.object},
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse(
            "dashboard:shipping-method-detail", kwargs={"pk": self.object.pk}
        )


class WeightBasedDetailView(generic.CreateView):
    model = WeightBand
    form_class = WeightBandForm
    template_name = "oscar/dashboard/shipping/weight_based_detail.html"

    # pylint: disable=attribute-defined-outside-init
    def dispatch(self, request, *args, **kwargs):
        self.method = shortcuts.get_object_or_404(WeightBased, pk=kwargs["pk"])
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self, **kwargs):
        kwargs = super().get_form_kwargs(**kwargs)
        kwargs["method"] = self.method
        return kwargs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["method"] = self.method
        return ctx

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/band_created.html", {"band": self.object}
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse(
            "dashboard:shipping-method-detail", kwargs={"pk": self.method.pk}
        )


class WeightBasedUpdateView(generic.UpdateView):
    model = WeightBased
    form_class = WeightBasedForm
    template_name = "oscar/dashboard/shipping/weight_based_form.html"
    context_object_name = "method"

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/method_updated.html",
            {"method": self.object},
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse(
            "dashboard:shipping-method-detail", kwargs={"pk": self.object.pk}
        )


class WeightBandUpdateView(generic.UpdateView):
    model = WeightBand
    form_class = WeightBandForm
    template_name = "oscar/dashboard/shipping/weight_band_form.html"
    context_object_name = "band"

    # pylint: disable=attribute-defined-outside-init
    def dispatch(self, request, *args, **kwargs):
        self.method = shortcuts.get_object_or_404(WeightBased, pk=kwargs["method_pk"])
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.method.bands.all()

    def get_form_kwargs(self, **kwargs):
        kwargs = super().get_form_kwargs(**kwargs)
        kwargs["method"] = self.method
        return kwargs

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/band_updated.html", {"band": self.object}
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse(
            "dashboard:shipping-method-detail", kwargs={"pk": self.method.pk}
        )


class WeightBandDeleteView(generic.DeleteView):
    model = WeightBased
    template_name = "oscar/dashboard/shipping/weight_band_delete.html"
    context_object_name = "band"

    # pylint: disable=attribute-defined-outside-init
    def dispatch(self, request, *args, **kwargs):
        self.method = shortcuts.get_object_or_404(WeightBased, pk=kwargs["method_pk"])
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.method.bands.all()

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/band_deleted.html", {"band": self.object}
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse(
            "dashboard:shipping-method-detail", kwargs={"pk": self.method.pk}
        )


class WeightBasedDeleteView(generic.DeleteView):
    model = WeightBased
    template_name = "oscar/dashboard/shipping/weight_based_delete.html"
    context_object_name = "method"

    def get_success_url(self):
        msg = render_to_string(
            "oscar/dashboard/shipping/messages/method_deleted.html",
            {"method": self.object},
        )
        messages.success(self.request, msg, extra_tags="safe noicon")
        return reverse("dashboard:shipping-method-list")



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class CommunicationConfig(OscarConfig):
    label = "communication"
    name = "oscar.apps.communication"
    verbose_name = _("Communication")



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wishlists', '0002_auto_20160111_1108'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wishlist',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date created'),
        ),
    ]



from oscar.apps.voucher.abstract_models import (
    AbstractVoucher,
    AbstractVoucherApplication,
    AbstractVoucherSet,
)
from oscar.core.loading import is_model_registered

__all__ = []

if not is_model_registered("voucher", "VoucherSet"):

    class VoucherSet(AbstractVoucherSet):
        pass

    __all__.append("VoucherSet")


if not is_model_registered("voucher", "Voucher"):

    class Voucher(AbstractVoucher):
        pass

    __all__.append("Voucher")


if not is_model_registered("voucher", "VoucherApplication"):

    class VoucherApplication(AbstractVoucherApplication):
        pass

    __all__.append("VoucherApplication")



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class ReportsDashboardConfig(OscarDashboardConfig):
    label = "reports_dashboard"
    name = "oscar.apps.dashboard.reports"
    verbose_name = _("Reports dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "reports-index": DashboardPermission.get("user_record"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.index_view = get_class("dashboard.reports.views", "IndexView")
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.index_view.as_view(), name="reports-index"),
        ]
        return self.post_process_urls(urls)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-13 08:44
from __future__ import unicode_literals

from django.db import migrations
import oscar.models.fields.slugfield


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0008_auto_20160304_1652'),
    ]

    operations = [
        migrations.AlterField(
            model_name='category',
            name='slug',
            field=oscar.models.fields.slugfield.SlugField(max_length=255, verbose_name='Slug'),
        ),
    ]



# -*- coding: utf-8 -*-
from django.conf import settings
from django.db import connection
from django.db.models.signals import post_delete, post_save, m2m_changed
from django.dispatch import receiver

from oscar.core.loading import get_model
from oscar.checks import use_productcategory_materialised_view

Category = get_model("catalogue", "Category")
ProductCategory = get_model("catalogue", "ProductCategory")


if settings.OSCAR_DELETE_IMAGE_FILES:
    from django.db import models

    from oscar.core.thumbnails import get_thumbnailer

    ProductImage = get_model("catalogue", "ProductImage")

    # pylint: disable=unused-argument
    def delete_image_files(sender, instance, **kwargs):
        """
        Deletes the original image and created thumbnails.
        """
        image_fields = (models.ImageField,)
        thumbnailer = get_thumbnailer()
        for field in instance._meta.fields:
            if isinstance(field, image_fields):
                # Make Django return ImageFieldFile instead of ImageField
                field_file = getattr(instance, field.name)
                thumbnailer.delete_thumbnails(field_file)

    # Connect for all models with ImageFields - add as needed
    models_with_images = [ProductImage, Category]
    for image_instance in models_with_images:
        post_delete.connect(delete_image_files, sender=image_instance)


# pylint: disable=unused-argument
@receiver(post_save, sender=Category, dispatch_uid="set_ancestors_are_public")
def post_save_set_ancestors_are_public(sender, instance, **kwargs):
    if kwargs.get("raw"):
        return

    instance.set_ancestors_are_public()


@receiver([post_save, post_delete, m2m_changed], sender=ProductCategory)
def refresh_materialized_view(sender, **kwargs):
    if kwargs.get("raw") or not use_productcategory_materialised_view():
        return

    with connection.cursor() as cursor:
        cursor.execute(
            "REFRESH MATERIALIZED VIEW CONCURRENTLY catalogue_product_category_hierarchy;"
        )



# Generated by Django 3.1.2 on 2021-02-24 07:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0009_make_voucher_names_unique'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voucher',
            name='name',
            field=models.CharField(help_text='This will be shown in the checkout and basket once the voucher is entered', max_length=128, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='voucherset',
            name='name',
            field=models.CharField(max_length=100, unique=True, verbose_name='Name'),
        ),
        migrations.RemoveField(
            model_name='voucherset',
            name='offer',
        ),
    ]



import warnings

from oscar.utils.deprecation import RemovedInOscar32Warning


def deprecated(obj):
    if isinstance(obj, type):
        return _deprecated_cls(cls=obj)
    else:
        return _deprecated_func(f=obj)


def _deprecated_func(f, warn_cls=RemovedInOscar32Warning):
    def _deprecated(*args, **kwargs):
        message = (
            "Method '%s' is deprecated and will be "
            "removed in the next version of django-oscar" % f.__name__
        )
        warnings.warn(message, warn_cls, stacklevel=2)
        return f(*args, **kwargs)

    return _deprecated


def _deprecated_cls(cls, warn_cls=RemovedInOscar32Warning):
    class Deprecated(cls):
        def __init__(self, *args, **kwargs):
            message = (
                "Class '%s' is deprecated and will be "
                "removed in the next version of django-oscar" % cls.__name__
            )
            warnings.warn(message, warn_cls, stacklevel=2)
            super().__init__(*args, **kwargs)

    return Deprecated



# Generated by Django 5.1.4 on 2025-02-17 12:47

from django.db import migrations, connection

from oscar.checks import is_postgres

def create_materialized_view(apps, schema_editor):
    """Creates a materialized view for PostgreSQL."""
    if is_postgres():
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE MATERIALIZED VIEW IF NOT EXISTS catalogue_product_category_hierarchy AS
                WITH RECURSIVE category_hierarchy AS (
                    SELECT id, path FROM catalogue_category
                )
                SELECT DISTINCT 
                    CONCAT(p.id, '-', parent_categories.id) AS id,
                    p.id AS product_id, 
                    parent_categories.id AS category_id
                FROM catalogue_productcategory pc
                JOIN catalogue_category child_categories ON pc.category_id = child_categories.id
                JOIN catalogue_category parent_categories 
                    ON child_categories.path LIKE parent_categories.path || '%'
                JOIN catalogue_product p ON p.id = pc.product_id;
                
            """)
            
            cursor.execute("""
                CREATE UNIQUE INDEX CONCURRENTLY IF NOT EXISTS catalogue_product_category_hierarchy_idx 
                ON catalogue_product_category_hierarchy (product_id, category_id);
            """)


def drop_materialized_view(apps, schema_editor):
    """Drops the materialized view if rolling back."""
    if is_postgres():
        with connection.cursor() as cursor:
            cursor.execute("DROP MATERIALIZED VIEW IF EXISTS catalogue_product_category_hierarchy;")
            cursor.execute("DROP INDEX IF EXISTS catalogue_product_category_hierarchy_idx;")


class Migration(migrations.Migration):
    atomic = False
    
    dependencies = [
        ("catalogue", "0029_product_code"),
    ]

    operations = [
         migrations.RunPython(create_materialized_view, drop_materialized_view),
     ]



import json

from django.contrib import messages
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView

from oscar.core.loading import get_model

ConditionalOffer = get_model("offer", "ConditionalOffer")


class OfferWizardStepView(FormView):
    wizard_name = "offer_wizard"
    form_class = None
    step_name = None
    update = False
    url_name = None
    success_url_name = "dashboard:offer-list"

    # Keep a reference to previous view class to allow checks to be made on
    # whether prior steps have been completed
    previous_view = None

    # pylint: disable=attribute-defined-outside-init
    def dispatch(self, request, *args, **kwargs):
        if self.update:
            self.offer = get_object_or_404(ConditionalOffer, id=kwargs["pk"])
        if not self.is_previous_step_complete(request):
            messages.warning(
                request,
                _("%s step not complete") % (self.previous_view.step_name.title(),),
            )
            return HttpResponseRedirect(reverse(self.previous_view.url_name))
        return super().dispatch(request, *args, **kwargs)

    def is_previous_step_complete(self, request):
        if not self.previous_view:
            return True
        return self.previous_view.is_valid(self, request)

    def _key(self, step_name=None, is_object=False):
        key = step_name if step_name else self.step_name
        if self.update:
            key += str(self.offer.id)
        if is_object:
            key += "_obj"
        return key

    def _store_form_kwargs(self, form):
        session_data = self.request.session.setdefault(self.wizard_name, {})

        # Adjust kwargs to avoid trying to save the range instance
        form_data = form.cleaned_data.copy()
        product_range = form_data.get("range")
        if product_range is not None:
            form_data["range"] = product_range.id

        form_kwargs = {"data": form_data}
        json_data = json.dumps(form_kwargs, cls=DjangoJSONEncoder)

        session_data[self._key()] = json_data
        self.request.session.save()

    def _fetch_form_kwargs(self):
        session_data = self.request.session.setdefault(self.wizard_name, {})
        json_data = session_data.get(self._key(self.step_name), None)
        if json_data:
            return json.loads(json_data)

        return {}

    def _store_object(self, form):
        session_data = self.request.session.setdefault(self.wizard_name, {})

        # We don't store the object instance as that is not JSON serialisable.
        # Instead, we save an alternative form
        instance = form.save(commit=False)
        # remove fields that do not exist (yet) on the uncommitted instance, i.e. m2m fields
        cleanfields = [field.name for field in instance._meta.local_fields]

        json_qs = serializers.serialize("json", [instance], fields=tuple(cleanfields))

        session_data[self._key(is_object=True)] = json_qs
        self.request.session.save()

    def _fetch_object(self, step_name, request=None):
        if request is None:
            request = self.request
        session_data = request.session.setdefault(self.wizard_name, {})
        json_qs = session_data.get(self._key(step_name, is_object=True), None)
        if json_qs:
            # Recreate model instance from passed data
            deserialised_obj = list(serializers.deserialize("json", json_qs))
            return deserialised_obj[0].object

    def _fetch_session_offer(self):
        """Return the offer instance loaded with the data stored in the session.

        When updating an offer, the updated fields are used with the existing
        offer data.
        """
        offer = self._fetch_object("metadata")
        if offer is None and self.update:
            offer = self.offer
        return offer

    def _flush_session(self):
        self.request.session[self.wizard_name] = {}
        self.request.session.save()

    def get_form_kwargs(self, *args, **kwargs):
        form_kwargs = {}
        if self.update:
            form_kwargs["instance"] = self.get_instance()
        session_kwargs = self._fetch_form_kwargs()
        form_kwargs.update(session_kwargs)
        parent_kwargs = super().get_form_kwargs(*args, **kwargs)
        form_kwargs.update(parent_kwargs)
        return form_kwargs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        if self.update:
            ctx["offer"] = self.offer
        ctx["session_offer"] = self._fetch_session_offer()
        ctx["title"] = self.get_title()
        return ctx

    def get_title(self):
        return self.step_name.title()

    def form_valid(self, form):
        self._store_form_kwargs(form)
        self._store_object(form)

        if self.update and "save" in form.data:
            # Save changes to this offer when updating and pressed save button
            return self.save_offer(self.offer)
        else:
            # Proceed to next page
            return super().form_valid(form)

    def save_offer(self, offer):
        # We update the offer with the name/description/offer_type from step 1
        session_offer = self._fetch_session_offer()
        offer.name = session_offer.name
        offer.description = session_offer.description
        offer.offer_type = session_offer.offer_type

        # Save the related models, then save the offer.
        # Note than you can save already on the first page of the wizard,
        # so le'ts check if the benefit and condition exist
        benefit = self._fetch_object("benefit")
        if benefit:
            benefit.save()
            offer.benefit = benefit

        condition = self._fetch_object("condition")
        if condition:
            condition.save()
            offer.condition = condition

        offer.save()

        self._flush_session()

        if self.update:
            msg = _("Offer '%s' updated") % offer.name
        else:
            msg = _("Offer '%s' created!") % offer.name
        messages.success(self.request, msg)

        return HttpResponseRedirect(
            reverse("dashboard:offer-detail", kwargs={"pk": offer.pk})
        )

    def get_success_url(self):
        if self.update:
            return reverse(self.success_url_name, kwargs={"pk": self.kwargs["pk"]})
        return reverse(self.success_url_name)

    @classmethod
    def is_valid(cls, current_view, request):
        if current_view.update:
            return True
        return current_view._fetch_object(cls.step_name, request) is not None



# Generated by Django 3.1.2 on 2021-02-24 06:18

from django.db import migrations


def make_voucher_names_unique(apps, schema_editor):
    """
    Appends a number to non-unique voucher names.
    """
    Voucher = apps.get_model('voucher', 'Voucher')
    vouchers = Voucher.objects.order_by('date_created')
    # Find vouchers with non-unique names
    vouchers_for_name = {}
    for voucher in vouchers:
        if voucher.name not in vouchers_for_name:
            vouchers_for_name[voucher.name] = []
        vouchers_for_name[voucher.name].append(voucher.id)
    # Change names for vouchers with non-unique names
    for voucher in vouchers:
        if len(vouchers_for_name[voucher.name]) > 1:
            voucher.name = "%s - %d" % (voucher.name, vouchers_for_name[voucher.name].index(voucher.id) + 1)
            voucher.save()


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0008_auto_20200801_0817'),
    ]

    operations = [
        migrations.RunPython(make_voucher_names_unique, migrations.RunPython.noop),
    ]



# Generated by Django 5.1.4 on 2025-02-17 14:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0030_auto_20250217_1247"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProductCategoryHierarchy",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=255,
                        primary_key=True,
                    ),
                ),
                ("product_id", models.IntegerField()),
                ("category_id", models.IntegerField()),
            ],
            options={
                "verbose_name": "Product category hierarchy",
                "verbose_name_plural": "Product categories hierarchy",
                "db_table": "catalogue_product_category_hierarchy",
                "managed": False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.core.validators
import oscar.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0005_auto_20150604_1450'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productattribute',
            name='code',
            field=models.SlugField(max_length=128, verbose_name='Code', validators=[django.core.validators.RegexValidator(regex=r'^[a-zA-Z_][0-9a-zA-Z_]*$', message="Code can only contain the letters a-z, A-Z, digits, and underscores, and can't start with a digit"), oscar.core.validators.non_python_keyword]),
            preserve_default=True,
        ),
    ]



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('address', '0005_regenerate_user_address_hashes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='country',
            name='printable_name',
            field=models.CharField(db_index=True, max_length=128, verbose_name='Country name'),
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-10 05:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0021_auto_20201005_0844'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_boolean',
            field=models.BooleanField(blank=True, db_index=True, null=True, verbose_name='Boolean'),
        ),
    ]



# Generated by Django 2.0.7 on 2018-11-15 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0007_conditionaloffer_exclusive'),
    ]

    operations = [
        migrations.AlterField(
            model_name='conditionaloffer',
            name='priority',
            field=models.IntegerField(db_index=True, default=0, help_text='The highest priority offers are applied first', verbose_name='Priority'),
        ),
        migrations.AlterField(
            model_name='rangeproductfileupload',
            name='date_uploaded',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date Uploaded'),
        ),
    ]



from haystack import indexes

from oscar.core.loading import get_class, get_model

# Load default strategy (without a user/request)
is_solr_supported = get_class("search.features", "is_solr_supported")
Selector = get_class("partner.strategy", "Selector")


class ProductIndex(indexes.SearchIndex, indexes.Indexable):
    # Search text
    text = indexes.CharField(
        document=True,
        use_template=True,
        template_name="oscar/search/indexes/product/item_text.txt",
    )

    upc = indexes.CharField(model_attr="upc", null=True)
    title = indexes.EdgeNgramField(model_attr="title", null=True)
    title_exact = indexes.CharField(model_attr="title", null=True, indexed=False)

    # Fields for faceting
    product_class = indexes.CharField(null=True, faceted=False)
    category = indexes.MultiValueField(null=True, faceted=True)
    price = indexes.FloatField(null=True, faceted=True)
    num_in_stock = indexes.IntegerField(null=True, faceted=True)
    rating = indexes.IntegerField(null=True, faceted=True)

    # Spelling suggestions
    suggestions = indexes.FacetCharField()

    date_created = indexes.DateTimeField(model_attr="date_created")
    date_updated = indexes.DateTimeField(model_attr="date_updated")

    is_public = indexes.BooleanField(model_attr="is_public")
    structure = indexes.CharField(model_attr="structure")

    _strategy = None

    def get_model(self):
        return get_model("catalogue", "Product")

    def index_queryset(self, using=None):
        # Only index browsable products (not each individual child product)
        return self.get_model().objects.browsable().order_by("-date_updated")

    def read_queryset(self, using=None):
        return self.get_model().objects.browsable().base_queryset()

    def prepare_product_class(self, obj):
        return obj.get_product_class().name

    def prepare_category(self, obj):
        return list(obj.get_categories().values_list("pk", flat=True)) or []

    def prepare_rating(self, obj):
        if obj.rating is not None:
            return int(obj.rating)

    # Pricing and stock is tricky as it can vary per customer.  However, the
    # most common case is for customers to see the same prices and stock levels
    # and so we implement that case here.

    def get_strategy(self):
        if not self._strategy:
            self._strategy = Selector().strategy()
        return self._strategy

    def prepare_price(self, obj):
        strategy = self.get_strategy()
        result = None
        if obj.is_parent:
            result = strategy.fetch_for_parent(obj)
        elif obj.has_stockrecords:
            result = strategy.fetch_for_product(obj)

        if result:
            if result.price.is_tax_known:
                return result.price.incl_tax
            return result.price.excl_tax

    def prepare_num_in_stock(self, obj):
        strategy = self.get_strategy()
        if obj.is_parent:
            # Don't return a stock level for parent products
            return None
        elif obj.has_stockrecords:
            result = strategy.fetch_for_product(obj)
            return result.stockrecord.net_stock_level

    def prepare(self, obj):
        prepared_data = super().prepare(obj)

        # We use Haystack's dynamic fields to ensure that the title field used
        # for sorting is of type "string'.
        if is_solr_supported():
            prepared_data["title_s"] = prepared_data["title"]

        # Use title for spelling suggestions
        prepared_data["suggestions"] = prepared_data["title"]

        return prepared_data

    def get_updated_field(self):
        """
        Used to specify the field used to determine if an object has been
        updated

        Can be used to filter the query set when updating the index
        """
        return "date_updated"



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-29 08:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('reviews', '0003_auto_20160802_1358'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productreview',
            name='product',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='reviews', to='catalogue.Product'),
        ),
    ]



# Generated by Django 2.2.4 on 2019-08-16 08:38

from django.db import migrations, models
import oscar.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0016_auto_20190327_0757'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_file',
            field=models.FileField(blank=True, max_length=255, null=True, upload_to=oscar.utils.models.get_image_upload_path),
        ),
        migrations.AlterField(
            model_name='productattributevalue',
            name='value_image',
            field=models.ImageField(blank=True, max_length=255, null=True, upload_to=oscar.utils.models.get_image_upload_path),
        ),
        migrations.AlterField(
            model_name='productimage',
            name='original',
            field=models.ImageField(max_length=255, upload_to=oscar.utils.models.get_image_upload_path, verbose_name='Original'),
        ),
    ]



import phonenumbers
from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _
from phonenumber_field.phonenumber import PhoneNumber


class PhoneNumberMixin(object):
    """Validation mixin for forms with a phone numbers, and optionally a country.

    It tries to validate the phone numbers, and on failure tries to validate
    them using a hint (the country provided), and treating it as a local number.

    Specify which fields to treat as phone numbers by specifying them in
    `phone_number_fields`, a dictionary of fields names and default kwargs
    for instantiation of the field.
    """

    country = None
    region_code = None
    # Since this mixin will be used with `ModelForms`, names of phone number
    # fields should match names of the related Model field
    phone_number_fields = {
        "phone_number": {
            "required": False,
            "help_text": "",
            "max_length": 32,
            "label": _("Phone number"),
        },
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We can't use the PhoneNumberField here since we want validate the
        # phonenumber based on the selected country as a fallback when a local
        # number is entered. We add the fields in the init since on Python 2
        # using forms.Form as base class results in errors when using this
        # class as mixin.

        # If the model field already exists, copy existing properties from it
        for field_name, field_kwargs in self.phone_number_fields.items():
            for key in field_kwargs:
                try:
                    field_kwargs[key] = getattr(self.fields[field_name], key)
                except (KeyError, AttributeError):
                    pass

            self.fields[field_name] = forms.CharField(**field_kwargs)

    def get_country(self):
        # If the form data contains valid country information, we use that.
        if hasattr(self, "cleaned_data") and "country" in self.cleaned_data:
            return self.cleaned_data["country"]
        # Oscar hides the field if there's only one country. Then (and only
        # then!) we can consider a country on the model instance.
        elif "country" not in self.fields and hasattr(self.instance, "country"):
            return self.instance.country

    def set_country_and_region_code(self):
        # Try hinting with the shipping country if we can determine one.
        self.country = self.get_country()
        if self.country:
            self.region_code = self.country.iso_3166_1_a2

    def clean_phone_number_field(self, field_name):
        number = self.cleaned_data.get(field_name)

        # Empty
        if number in validators.EMPTY_VALUES:
            return ""

        # Check for an international phone format
        try:
            phone_number = PhoneNumber.from_string(number)
        except phonenumbers.NumberParseException:
            if not self.region_code:
                # There is no shipping country, not a valid international number
                self.add_error(
                    field_name, _("This is not a valid international phone format.")
                )
                return number

            # The PhoneNumber class does not allow specifying
            # the region. So we drop down to the underlying phonenumbers
            # library, which luckily allows parsing into a PhoneNumber
            # instance.
            try:
                phone_number = PhoneNumber.from_string(number, region=self.region_code)
                if not phone_number.is_valid():
                    self.add_error(
                        field_name,
                        _("This is not a valid local phone format for %s.")
                        % self.country,
                    )
            except phonenumbers.NumberParseException:
                # Not a valid local or international phone number
                self.add_error(
                    field_name,
                    _("This is not a valid local or international phone format."),
                )
                return number

        return phone_number

    def clean(self):
        self.set_country_and_region_code()
        cleaned_data = super().clean()
        for field_name in self.phone_number_fields:
            cleaned_data[field_name] = self.clean_phone_number_field(field_name)
        return cleaned_data



from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig


class AnalyticsConfig(OscarConfig):
    label = "analytics"
    name = "oscar.apps.analytics"
    verbose_name = _("Analytics")

    # pylint: disable=unused-import
    def ready(self):
        from . import receivers



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='category',
            options={'ordering': ['path'], 'verbose_name': 'Category', 'verbose_name_plural': 'Categories'},
        ),
        migrations.RemoveField(
            model_name='category',
            name='full_name',
        ),
    ]



from django.contrib import admin

from oscar.core.loading import get_model

ConditionalOffer = get_model("offer", "ConditionalOffer")
Condition = get_model("offer", "Condition")
Benefit = get_model("offer", "Benefit")
Range = get_model("offer", "Range")


class ConditionAdmin(admin.ModelAdmin):
    list_display = ("type", "value", "range")


class BenefitAdmin(admin.ModelAdmin):
    list_display = ("__str__", "type", "value", "range")


class ConditionalOfferAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "offer_type",
        "start_datetime",
        "end_datetime",
        "condition",
        "benefit",
        "total_discount",
    )
    list_filter = ("offer_type",)
    readonly_fields = ("total_discount", "num_orders")
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "description",
                    "offer_type",
                    "condition",
                    "benefit",
                    "start_datetime",
                    "end_datetime",
                    "priority",
                )
            },
        ),
        ("Usage", {"fields": ("total_discount", "num_orders")}),
    )


admin.site.register(ConditionalOffer, ConditionalOfferAdmin)
admin.site.register(Condition, ConditionAdmin)
admin.site.register(Benefit, BenefitAdmin)
admin.site.register(Range)



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class OffersDashboardConfig(OscarDashboardConfig):
    label = "offers_dashboard"
    name = "oscar.apps.dashboard.offers"
    verbose_name = _("Offers dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "offer-list": DashboardPermission.get("offer"),
            "offer-metadata": DashboardPermission.get("offer"),
            "offer-condition": DashboardPermission.get("offer"),
            "offer-benefit": DashboardPermission.get("offer"),
            "offer-restrictions": DashboardPermission.get("offer"),
            "offer-delete": DashboardPermission.get("offer"),
            "offer-detail": DashboardPermission.get("offer"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.offers.views", "OfferListView")
        self.metadata_view = get_class("dashboard.offers.views", "OfferMetaDataView")
        self.condition_view = get_class("dashboard.offers.views", "OfferConditionView")
        self.benefit_view = get_class("dashboard.offers.views", "OfferBenefitView")
        self.restrictions_view = get_class(
            "dashboard.offers.views", "OfferRestrictionsView"
        )
        self.delete_view = get_class("dashboard.offers.views", "OfferDeleteView")
        self.detail_view = get_class("dashboard.offers.views", "OfferDetailView")
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="offer-list"),
            # Creation
            path("new/metadata/", self.metadata_view.as_view(), name="offer-metadata"),
            path(
                "new/condition/", self.condition_view.as_view(), name="offer-condition"
            ),
            path("new/incentive/", self.benefit_view.as_view(), name="offer-benefit"),
            path(
                "new/restrictions/",
                self.restrictions_view.as_view(),
                name="offer-restrictions",
            ),
            # Update
            path(
                "<int:pk>/metadata/",
                self.metadata_view.as_view(update=True),
                name="offer-metadata",
            ),
            path(
                "<int:pk>/condition/",
                self.condition_view.as_view(update=True),
                name="offer-condition",
            ),
            path(
                "<int:pk>/incentive/",
                self.benefit_view.as_view(update=True),
                name="offer-benefit",
            ),
            path(
                "<int:pk>/restrictions/",
                self.restrictions_view.as_view(update=True),
                name="offer-restrictions",
            ),
            # Delete
            path("<int:pk>/delete/", self.delete_view.as_view(), name="offer-delete"),
            # Stats
            path("<int:pk>/", self.detail_view.as_view(), name="offer-detail"),
        ]
        return self.post_process_urls(urls)



# pylint: disable=W0201
from django.contrib.auth.decorators import login_required
from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class BasketConfig(OscarConfig):
    label = "basket"
    name = "oscar.apps.basket"
    verbose_name = _("Basket")

    namespace = "basket"

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.summary_view = get_class("basket.views", "BasketView")
        self.saved_view = get_class("basket.views", "SavedView")
        self.add_view = get_class("basket.views", "BasketAddView")
        self.add_voucher_view = get_class("basket.views", "VoucherAddView")
        self.remove_voucher_view = get_class("basket.views", "VoucherRemoveView")

    def get_urls(self):
        urls = [
            path("", self.summary_view.as_view(), name="summary"),
            path("add/<int:pk>/", self.add_view.as_view(), name="add"),
            path("vouchers/add/", self.add_voucher_view.as_view(), name="vouchers-add"),
            path(
                "vouchers/<int:pk>/remove/",
                self.remove_voucher_view.as_view(),
                name="vouchers-remove",
            ),
            path("saved/", login_required(self.saved_view.as_view()), name="saved"),
        ]
        return self.post_process_urls(urls)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.db import models, migrations
import oscar.models.fields.autoslugfield
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CommunicationEventType',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Code', editable=False, separator='_', max_length=128, help_text='Code used for looking up this event programmatically', blank=True)),
                ('name', models.CharField(verbose_name='Name', max_length=255, help_text='This is just used for organisational purposes')),
                ('category', models.CharField(default='Order related', max_length=255, verbose_name='Category', choices=[('Order related', 'Order related'), ('User related', 'User related')])),
                ('email_subject_template', models.CharField(verbose_name='Email Subject Template', max_length=255, blank=True, null=True)),
                ('email_body_template', models.TextField(blank=True, verbose_name='Email Body Template', null=True)),
                ('email_body_html_template', models.TextField(verbose_name='Email Body HTML Template', blank=True, help_text='HTML template', null=True)),
                ('sms_template', models.CharField(verbose_name='SMS Template', max_length=170, help_text='SMS template', blank=True, null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
            ],
            options={
                'verbose_name_plural': 'Communication event types',
                'verbose_name': 'Communication event type',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.TextField(max_length=255, verbose_name='Subject')),
                ('body_text', models.TextField(verbose_name='Body Text')),
                ('body_html', models.TextField(verbose_name='Body HTML', blank=True)),
                ('date_sent', models.DateTimeField(auto_now_add=True, verbose_name='Date Sent')),
                ('user', models.ForeignKey(verbose_name='User', related_name='emails', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Emails',
                'verbose_name': 'Email',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('category', models.CharField(max_length=255, blank=True)),
                ('location', models.CharField(default='Inbox', max_length=32, choices=[('Inbox', 'Inbox'), ('Archive', 'Archive')])),
                ('date_sent', models.DateTimeField(auto_now_add=True)),
                ('date_read', models.DateTimeField(blank=True, null=True)),
                ('recipient', models.ForeignKey(related_name='notifications', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
                ('sender', models.ForeignKey(to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('-date_sent',),
                'verbose_name_plural': 'Notifications',
                'verbose_name': 'Notification',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ProductAlert',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=75, db_index=True, verbose_name='Email', blank=True)),
                ('key', models.CharField(max_length=128, db_index=True, verbose_name='Key', blank=True)),
                ('status', models.CharField(default='Active', max_length=20, verbose_name='Status', choices=[('Unconfirmed', 'Not yet confirmed'), ('Active', 'Active'), ('Cancelled', 'Cancelled'), ('Closed', 'Closed')])),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('date_confirmed', models.DateTimeField(blank=True, verbose_name='Date confirmed', null=True)),
                ('date_cancelled', models.DateTimeField(blank=True, verbose_name='Date cancelled', null=True)),
                ('date_closed', models.DateTimeField(blank=True, verbose_name='Date closed', null=True)),
                ('product', models.ForeignKey(to='catalogue.Product', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(null=True, verbose_name='User', related_name='alerts', to=settings.AUTH_USER_MODEL, blank=True, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Product alerts',
                'verbose_name': 'Product alert',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
    ]



from oscar.core.loading import get_class

FacetedSearchView = get_class("search.views.search", "FacetedSearchView")
CatalogueView = get_class("search.views.catalogue", "CatalogueView")
ProductCategoryView = get_class("search.views.catalogue", "ProductCategoryView")



from decimal import Decimal as D


class OfferApplications(object):
    """
    A collection of offer applications and the discounts that they give.

    Each offer application is stored as a dict which has fields for:

    * The offer that led to the successful application
    * The result instance
    * The number of times the offer was successfully applied
    """

    def __init__(self):
        self.applications = {}

    def __iter__(self):
        return self.applications.values().__iter__()

    def __len__(self):
        return len(self.applications)

    def add(self, offer, result):
        if offer.id not in self.applications:
            self.applications[offer.id] = {
                "offer": offer,
                "result": result,
                "name": offer.name,
                "description": result.description,
                "voucher": offer.get_voucher(),
                "freq": 0,
                "discount": D("0.00"),
            }
        self.applications[offer.id]["discount"] += result.discount
        self.applications[offer.id]["freq"] += 1

    @property
    def offer_discounts(self):
        """
        Return basket discounts from offers (but not voucher offers)
        """
        discounts = []
        for application in self.applications.values():
            if not application["voucher"] and application["discount"] > 0:
                discounts.append(application)
        return discounts

    @property
    def voucher_discounts(self):
        """
        Return basket discounts from vouchers.
        """
        discounts = []
        for application in self.applications.values():
            if application["voucher"] and application["discount"] > 0:
                discounts.append(application)
        return discounts

    @property
    def shipping_discounts(self):
        """
        Return shipping discounts
        """
        discounts = []
        for application in self.applications.values():
            if application["result"].affects_shipping:
                discounts.append(application)
        return discounts

    @property
    def grouped_voucher_discounts(self):
        """
        Return voucher discounts aggregated up to the voucher level.

        This is different to the voucher_discounts property as a voucher can
        have multiple offers associated with it.
        """
        voucher_discounts = {}
        for application in self.voucher_discounts:
            voucher = application["voucher"]
            discount = application["discount"]
            if voucher.code not in voucher_discounts:
                voucher_discounts[voucher.code] = {
                    "voucher": voucher,
                    "discount": discount,
                }
            else:
                voucher_discounts[voucher.code]["discount"] += discount
        return voucher_discounts.values()

    @property
    def post_order_actions(self):
        """
        Return successful offer applications which didn't lead to a discount
        """
        applications = []
        for application in self.applications.values():
            if application["result"].affects_post_order:
                applications.append(application)
        return applications

    @property
    def offers(self):
        """
        Return a dict of offers that were successfully applied
        """
        return dict([(a["offer"].id, a["offer"]) for a in self.applications.values()])


class ApplicationResult(object):
    is_final = is_successful = False
    # Basket discount
    discount = D("0.00")
    description = None

    # Offer applications can affect 3 distinct things
    # (a) Give a discount off the BASKET total
    # (b) Give a discount off the SHIPPING total
    # (a) Trigger a post-order action
    BASKET, SHIPPING, POST_ORDER = 0, 1, 2
    affects = None

    @property
    def affects_basket(self):
        return self.affects == self.BASKET

    @property
    def affects_shipping(self):
        return self.affects == self.SHIPPING

    @property
    def affects_post_order(self):
        return self.affects == self.POST_ORDER


class BasketDiscount(ApplicationResult):
    """
    For when an offer application leads to a simple discount off the basket's
    total
    """

    affects = ApplicationResult.BASKET

    def __init__(self, amount):
        self.discount = amount

    @property
    def is_successful(self):
        """
        Returns ``True`` if the discount is greater than zero
        """
        return self.discount > 0

    def __str__(self):
        return "<Basket discount of %s>" % self.discount

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.discount)


# Helper global as returning zero discount is quite common
ZERO_DISCOUNT = BasketDiscount(D("0.00"))


class ShippingDiscount(ApplicationResult):
    """
    For when an offer application leads to a discount from the shipping cost
    """

    is_successful = is_final = True
    affects = ApplicationResult.SHIPPING


SHIPPING_DISCOUNT = ShippingDiscount()


class PostOrderAction(ApplicationResult):
    """
    For when an offer condition is met but the benefit is deferred until after
    the order has been placed. E.g. buy 2 books and get 100 loyalty points.
    """

    is_final = is_successful = True
    affects = ApplicationResult.POST_ORDER

    def __init__(self, description):
        self.description = description
